"""Module containing several advanced plotting functions."""

from collections.abc import Sequence
from typing import Any, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from statannotations.Annotator import Annotator

from biopsykit.utils.functions import se

_PVALUE_THRESHOLDS = [[1e-3, "***"], [1e-2, "**"], [0.05, "*"]]


def lineplot(
    data: pd.DataFrame, x: str, y: str, hue: Optional[str] = None, style: Optional[str] = None, **kwargs
) -> tuple[plt.Figure, plt.Axes]:
    """Draw a line plot with error bars with the possibility of several semantic groupings.

    This is an extension to seaborn's lineplot function (:func:`seaborn.lineplot`).
    It offers the same interface, but several improvements:

    * Data points are not only connected as line, but are also drawn with marker.
    * Lines can have an offset along the categorical (x) axis for better visualization
      (seaborn equivalent: ``dodge``, which is only available for :func:`seaborn.pointplot`,
      not for :func:`seaborn.lineplot`).
    * Further plot parameters (axis labels, ticks, etc.) are inferred from the dataframe.

    Equivalent to seaborn, the relationship between ``x`` and ``y`` can be shown for different subsets of the
    data using the ``hue`` and ``style`` parameters. If both parameters are assigned two different grouping variables
    can be represented.

    Error bars are displayed as standard error.

    See the seaborn documentation for further information.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to plot
    x : str
        column of x axis in ``data``
    y : str
        column of y axis in ``data``
    hue : str, optional
        column name of grouping variable that will produce lines with different colors.
        Can be either categorical or numeric. If ``None`` then data will not be grouped.
    style : str, optional
        column name of grouping variable that will produce lines with different dashes and/or marker.
        If ``None`` then lines will not have different styles.
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``x_offset``: offset value to move different groups along the x axis for better visualization.
          Default: 0.05.
        * ``xlabel``: Label for x axis. If not specified it is inferred from the ``x`` column name.
        * ``ylabel``: Label for y axis. If not specified it is inferred from the ``y`` column name.
        * ``xticklabels``: List of labels for ticks of x axis. If not specified ``order`` is taken as tick labels.
          If ``order`` is not specified tick labels are inferred from x values.
        * ``ylim``: y-axis limits.
        * ``order``: list specifying the order of categorical values along the x axis.
        * ``hue_order``: list specifying the order of processing and plotting for categorical levels
          of the ``hue`` semantic.
        * ``marker``: string or list of strings to specify marker style.
          If ``marker`` is a string, then marker of each line will have the same style.
          If ``marker`` is a list, then marker of each line will have a different style.
        * ``linestyle``: string or list of strings to specify line style.
          If ``linestyle`` is a string, then each line will have the same style.
          If ``linestyle`` is a list, then each line will have a different style.
        * ``legend_fontsize``: font size of legend.
        * ``legend_loc``: location of legend in Axes.
        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.
        * ``err_kws``: additional parameters to control the aesthetics of the error bars.
          The ``err_kws`` are passed down to :meth:`matplotlib.axes.Axes.errorbar` or
          :meth:`matplotlib.axes.Axes.fill_between`, depending on ``err_style``. Parameters include:

          * ``capsize``: length of error bar caps in points


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`seaborn.lineplot`
        line plot function of Seaborn

    """
    x_offset = kwargs.get("x_offset", 0.01)
    multi_x_offset = kwargs.get("multi_x_offset", 0)
    order = kwargs.get("order")
    marker = kwargs.get("marker")
    linestyle = kwargs.get("linestyle")
    hue_order = kwargs.get("hue_order")
    palette = kwargs.get("palette")
    show_legend = kwargs.get("show_legend", True)
    if style is None:
        style = hue

    data = data.reset_index()

    fig, ax = _plot_get_fig_ax(**kwargs)

    kwargs.setdefault("ax", ax)
    kwargs.setdefault("err_kws", {"capsize": 5})

    marker, linestyle = _get_styles(data, style, hue, marker, linestyle)

    grouped = dict(tuple(data.groupby(hue))) if hue is not None else {y: data}

    if hue_order is not None:
        # reorder group dictionary
        grouped = {key: grouped[key] for key in hue_order}

    # generate x axis
    x_vals = data[x].unique()
    if all(isinstance(x, str) for x in x_vals):
        x_vals = np.arange(0, len(x_vals))
    span = x_vals[-1] - x_vals[0]
    x_offset = span * x_offset

    # iterate through groups
    for i, (key, df) in enumerate(grouped.items()):
        m_se = _get_df_lineplot(df, x, y, hue, order)

        err_kws = kwargs.get("err_kws", {})
        m = marker[i] if marker is not None else None
        ls = linestyle[i] if linestyle is not None else "-"
        c = palette[i] if palette is not None else None

        ax.errorbar(
            x=x_vals + multi_x_offset * span + x_offset * i,
            y=m_se["mean"],
            yerr=m_se["se"].values,
            marker=m,
            linestyle=ls,
            color=c,
            label=key,
            **err_kws,
        )

    _lineplot_style_axis(data, x, y, x_vals, order, **kwargs)

    if show_legend:
        _set_legend_errorbar(**kwargs)

    if kwargs.pop("tight_layout", True):
        fig.tight_layout()
    return fig, ax


def _lineplot_style_axis(data: pd.DataFrame, x: str, y: str, x_vals: np.ndarray, order: str, ax: plt.Axes, **kwargs):
    xlabel = kwargs.get("xlabel", x)
    ylabel = kwargs.get("ylabel", y)
    ylim = kwargs.get("ylim")

    if order is None:
        xticklabels = kwargs.get("xticklabels", data.groupby([x]).groups.keys())
    else:
        xticklabels = kwargs.get("xticklabels", order)

    ax.set_xticks(x_vals)
    ax.set_xticklabels(xticklabels)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_ylim(ylim)


def stacked_barchart(data: pd.DataFrame, **kwargs) -> tuple[plt.Figure, plt.Axes]:
    """Draw a stacked bar chart.

    A stacked bar chart has multiple bar charts along a categorical axis (``x`` axis) where values are
    stacked along the value axis (``y`` axis). The categorical axis corresponds to the columns in the dataframe
    whereas the value axis corresponds to the rows.

    This is an extension to the already existing function provided by pandas
    (``pandas.DataFrame.plot(kind='bar', stacked=True)``).


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to plot
    **kwargs
        Additional parameters to plotting function. For example, this can be:

        * ``order``: order of items along the categorical axis.
        * ``ylabel``: label of y axis.
        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.utils.dataframe_handling.stack_groups_percent`
        function to rearrange dataframe to be plotted as stacked bar chart

    """
    ax: plt.Axes = kwargs.get("ax")
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    ylabel = kwargs.get("ylabel")
    order = kwargs.get("order")
    if order:
        data = data.reindex(order)

    ax = data.plot(kind="bar", stacked=True, ax=ax, rot=0)
    ax.legend().set_title(None)
    if ylabel:
        ax.set_ylabel(ylabel)

    if kwargs.pop("tight_layout", True):
        fig.tight_layout()
    return fig, ax


def feature_boxplot(  # noqa: C901
    data: pd.DataFrame,
    x: Optional[str] = None,
    y: Optional[str] = None,
    order: Optional[Sequence[str]] = None,
    hue: Optional[str] = None,
    hue_order: Optional[Sequence[str]] = None,
    stats_kwargs: Optional[dict[str, Any]] = None,
    **kwargs,
) -> tuple[plt.Figure, plt.Axes]:
    """Draw boxplot with significance brackets.

    This is a wrapper of seaborn's boxplot function (:func:`~seaborn.boxplot`) that allows to add significance
    brackets that indicate statistical significance.


    Statistical annotations are plotted using ``statannot`` (https://github.com/webermarcolivier/statannot).
    This library can either use existing statistical results or perform statistical tests internally.
    To plot significance brackets a list of box pairs where annotations should be added are required.
    The p values can be provided as well, or, alternatively, be computed by ``statannot``.
    If :class:`~biopsykit.stats.StatsPipeline` was used for statistical analysis the list of box pairs and p values
    can be generated using :func:`~biopsykit.stats.StatsPipeline.sig_brackets` and passed in the ``stats_kws``
    parameter. Otherwise, see the ``statannot`` documentation  for a tutorial on how to specify significance brackets.

    .. note::
        The input data is assumed to be in long-format.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to plot
    x : str
        column of x axis in ``data``
    y : str
        column of y axis in ``data``
    order : list str, optional
        order to plot the categorical levels along the x axis.
    hue : str, optional
        column name of grouping variable. Default: ``None``
    hue_order : list of str, optional
        order to plot the grouping variable specified by ``hue``
    stats_kwargs : dict, optional
        dictionary with arguments for significance brackets.

        If annotations should be added, the following parameter is required:

        * ``box_pairs``: list of box pairs that should be annotated

        If already existing box pairs and p values should be used the following parameter is additionally required:

        * ``pvalues``: list of p values corresponding to ``box_pairs``

        If statistical tests should be computed by ``statsannot``, the following parameters are required:

        * ``test``: type of statistical test to be computed
        * ``comparisons_correction`` (optional): Whether (and which) type of multi-comparison correction should be
          applied. ``None`` to not apply any multi-comparison (default).

        The following parameters are optional:

        * ``pvalue_thresholds``: list of p value thresholds for statistical annotations. The default annotation is:
          '*': 0.01 <= p < 0.05, '**': 0.001 <= p < 0.01, `'***'`: p < 0.001
          (:math:`[[1e-3, "***"], [1e-2, "**"], [0.05, "*"]]`)

    **kwargs
        additional arguments that are passed down to :func:`~seaborn.boxplot`, for example:

        * ``ylabel``: label of y axis
        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~seaborn.boxplot`
        seaborn function to create boxplots
    :class:`~biopsykit.stats.StatsPipeline`
        class to create statistical analysis pipelines

    """
    fig, ax = _plot_get_fig_ax(**kwargs)
    kwargs.update({"ax": ax})

    if stats_kwargs is None:
        stats_kwargs = {}

    ylabel = kwargs.pop("ylabel", None)
    xticklabels = kwargs.pop("xticklabels", None)
    show_legend = kwargs.pop("show_legend", True)
    legend_fontsize = kwargs.pop("legend_fontsize", None)
    legend_loc = kwargs.pop("legend_loc", "upper right")
    legend_orientation = kwargs.pop("legend_orientation", "vertical")
    legend_title = kwargs.pop("legend_title", None)
    alpha = kwargs.pop("alpha", 1.0)
    kwargs.setdefault("saturation", 1.0)
    rect = kwargs.pop("rect", (0, 0, 0.825, 1.0) if legend_orientation == "vertical" else (0, 0, 1, 0.925))

    stats_kwargs = _feature_boxplot_sanitize_stats_kwargs(stats_kwargs)

    box_pairs = stats_kwargs.get("pairs", {})
    if len(box_pairs) == 0:
        stats_kwargs = {}

    palette = _sanitize_palette(data, kwargs.pop("palette", None), hue)
    if palette is not None:
        kwargs["palette"] = palette
    sns.boxplot(data=data.reset_index(), x=x, y=y, order=order, hue=hue, hue_order=hue_order, **kwargs)

    for patch in ax.patches:
        r, g, b, _ = patch.get_facecolor()
        patch.set_facecolor((r, g, b, alpha))

    if len(box_pairs) > 0:
        annotator = Annotator(
            data=data.reset_index(), ax=ax, x=x, y=y, order=order, hue=hue, hue_order=hue_order, **stats_kwargs
        )
        annotator.configure(hide_non_significant=True, pvalue_thresholds=stats_kwargs.get("pvalue_thresholds"))
        annotator.set_pvalues(stats_kwargs.get("pvalues"))
        annotator.annotate()

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if xticklabels is not None:
        ax.set_xticks(np.arange(len(xticklabels)))
        ax.set_xticklabels(xticklabels)

    if show_legend:
        if hue is not None and ax.legend_ is not None:
            ax.legend().remove()

        handles, labels = ax.get_legend_handles_labels()
        _feature_boxplot_add_legend(
            fig,
            hue,
            handles,
            labels,
            rect=rect,
            legend_loc=legend_loc,
            legend_fontsize=legend_fontsize,
            legend_orientation=legend_orientation,
            legend_title=legend_title,
        )
    elif kwargs.pop("tight_layout", True):
        fig.tight_layout()
    return fig, ax


def _feature_boxplot_sanitize_stats_kwargs(stats_kwargs: dict[str, Any]) -> dict[str, Any]:
    boxplot_pvals = stats_kwargs.get("pvalues", {})
    if len(stats_kwargs) > 0:
        stats_kwargs["comparisons_correction"] = stats_kwargs.get("comparisons_correction")
        stats_kwargs["test"] = stats_kwargs.get("test")

    if len(boxplot_pvals) > 0:
        stats_kwargs["perform_stat_test"] = False

    pairs = stats_kwargs.get("pairs", stats_kwargs.get("box_pairs"))
    stats_kwargs["pairs"] = pairs if pairs is not None else {}
    stats_kwargs.setdefault("pvalue_thresholds", _PVALUE_THRESHOLDS)
    return stats_kwargs


def multi_feature_boxplot(  # pylint:disable=too-many-branches
    data: pd.DataFrame,
    x: str,
    y: str,
    features: Union[Sequence[str], dict[str, Union[str, Sequence[str]]]],
    group: str,
    order: Optional[Sequence[str]] = None,
    hue: Optional[str] = None,
    hue_order: Optional[Sequence[str]] = None,
    stats_kwargs: Optional[dict] = None,
    **kwargs,
) -> tuple[plt.Figure, list[plt.Axes]]:
    """Draw multiple features as boxplots with significance brackets.

    For each feature, a new subplot will be created. Similarly to `feature_boxplot` subplots can be annotated
    with statistical significance brackets (can be specified via ``stats_kwargs`` parameter). For further information,
    see the :func:`~biopsykit.plotting.feature_boxplot` documentation.

    .. note::
        The input data is assumed to be in *long*-format.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to plot
    x : str
        column of x-axis in ``data``
    y : str
        column of y-axis in ``data``
    features : list of str or dict of str
        features to plot. If ``features`` is a list, each entry must correspond to one feature category in the
        index level specified by ``group``. A separate subplot will be created for each feature.
        If similar features (i.e., different `slope` or `AUC` parameters) should be combined into one subplot,
        ``features`` can be provided as dictionary.
        Then, the dict keys specify the feature category (a separate subplot will be created for each category)
        and the dict values specify the feature (or list of features) that are combined into the subplots.
    group : str
        name of index level with feature names. Corresponds to the subplots that are created.
    order : list of str, optional
        order to plot the categorical levels along the x axis
    hue : str, optional
        column name of grouping variable. Default: ``None``
    hue_order : list of str, optional
        order to plot the grouping variable specified by ``hue``
    stats_kwargs : dict, optional
        nested dictionary with arguments for significance brackets. The behavior and expected parameters are similar
        to the ``stats_kwargs`` parameter in :func:`~biopsykit.plotting.feature_boxplot`. However, the ``box_pairs``
        and ``pvalues`` arguments are expected not to be lists, but dictionaries with keys corresponding to the list
        entries (or the dict keys) in ``features`` and box pair / p-value lists are the dict values.
    **kwargs
        additional arguments that are passed down to :func:`~seaborn.boxplot`. For example:

        * ``order``: specifies x-axis order for subplots. Can be a list if order is the same for all subplots or a
          dict if order should be individual for subplots
        * ``xticklabels``: dictionary to set tick labels of x-axis in subplots. Keys correspond to the list entries
          (or the dict keys) in ``features``. Default: ``None``
        * ``ylabels``: dictionary to set y-axis labels in subplots. Keys correspond to the list entries
          (or the dict keys) in ``features``. Default: ``None``
        * ``axs``: list of pre-existing axes for the plot. Otherwise, a new figure and axes object is created and
          returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    axs : list of :class:`~matplotlib.axes.Axes`
        list of subplot axes objects


    See Also
    --------
    :func:`~seaborn.boxplot`
        seaborn function to create boxplots
    :class:`~biopsykit.plotting.feature_boxplot`
        plot single feature boxplot
    :class:`~biopsykit.stats.StatsPipeline`
        class to create statistical analysis pipelines and get parameter for plotting significance brackets

    """
    xlabels = kwargs.pop("xlabels", {})
    ylabels = kwargs.pop("ylabels", {})
    xticklabels = kwargs.pop("xticklabels", {})
    show_legend = kwargs.pop("show_legend", True)
    legend_fontsize = kwargs.pop("legend_fontsize", None)
    legend_title = kwargs.pop("legend_title", None)
    legend_loc = kwargs.pop("legend_loc", "upper right")
    legend_orientation = kwargs.pop("legend_orientation", "vertical")
    alpha = kwargs.pop("alpha", 1.0)
    kwargs.setdefault("saturation", 1.0)
    rect = kwargs.pop("rect", (0, 0, 0.825, 1.0) if legend_orientation == "vertical" else (0, 0, 1, 0.925))
    tight_layout = kwargs.pop("tight_layout", True)

    if isinstance(features, list):
        features = {f: [f] if isinstance(f, str) else f for f in features}

    axs: list[plt.Axes] = kwargs.pop("axs", kwargs.pop("ax", None))
    fig, axs = _plot_get_fig_ax_list(features, axs, **kwargs)

    if stats_kwargs is None:
        stats_kwargs = {}

    dict_box_pairs = stats_kwargs.pop("pairs", stats_kwargs.pop("box_pairs", None))
    dict_pvals = stats_kwargs.pop("pvalues", None)

    handles = None
    labels = None
    palette = kwargs.pop("palette", None)

    if hue is None:
        hue = x

    for ax, key in zip(axs, features):
        reindex_keys = features[key]
        if isinstance(reindex_keys, str):
            reindex_keys = [reindex_keys]
        data_plot = data.reindex(reindex_keys, level=group)
        if data_plot.empty:
            raise ValueError(f"Empty dataframe for '{key}'!")

        order_list = _multi_feature_boxplot_get_order(order, key)

        palette_hue = _sanitize_palette(data_plot, palette, hue)

        sns.boxplot(
            data=data_plot.reset_index(),
            x=x,
            y=y,
            order=order_list,
            hue=hue,
            legend=True,
            hue_order=hue_order,
            palette=palette_hue,
            ax=ax,
            **kwargs,
        )

        for patch in ax.patches:
            r, g, b, _ = patch.get_facecolor()
            patch.set_facecolor((r, g, b, alpha))

        _add_stat_annot_multi_feature_boxplot(
            data_plot,
            x,
            y,
            order_list,
            hue,
            hue_order,
            stats_kwargs,
            key,
            ax,
            dict_box_pairs,
            dict_pvals,
        )

        ax.set_ylabel(ylabels.get(key, ax.get_ylabel()))
        _style_xaxis_multi_feature_boxplot(ax, xlabels, xticklabels, key)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend().remove()

    if show_legend:
        _multi_feature_boxplot_add_legend(
            fig,
            hue,
            handles,
            labels,
            rect=rect,
            legend_loc=legend_loc,
            legend_fontsize=legend_fontsize,
            legend_orientation=legend_orientation,
            legend_title=legend_title,
        )
    if tight_layout:
        fig.tight_layout(rect=rect)
    return fig, axs


def _multi_feature_boxplot_get_order(order: Union[dict[str, Sequence[str]], Sequence[str]], key: str):
    if isinstance(order, dict):
        return order[key]
    return order


def feature_pairplot(
    data: pd.DataFrame, abbreviate_names: Optional[bool] = True, **kwargs
) -> Union[sns.PairGrid, tuple[sns.PairGrid, pd.DataFrame]]:
    """Plot feature pairs of a dataset.

    This function provides a convenient interface to the :func:`~seaborn.pairplot` class, with several additional
    features, such as showing abbreviated feature names in the plot with a description below the plot.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        DataFrame with feature data to plot. Data is expected to be in wide format.
    abbreviate_names : bool, optional
        if ``True``, feature names are abbreviated in the plot and a description is shown below the plot.
    **kwargs
        additional keyword arguments are passed down to :func:`~seaborn.pairplot`

    Returns
    -------
    :class:`~seaborn.FacetGrid`
        :class:`~seaborn.FacetGrid` object with the pairplot in it.

    See Also
    --------
    :class:`~seaborn.pairplot`
        seaborn class to create pairplots

    """
    if abbreviate_names:
        data = data.copy()
        feature_names = list(data.columns)
        # assign numbers to column names
        data.columns = range(len(data.columns))
        data.columns = data.columns.astype(str)
        data = data.add_prefix("feature #")

        # dataframe with assignment of column names to numbers
        feature_names_df = pd.DataFrame(feature_names, columns=["feature_name"])
        feature_names_df.index.name = "feature_id"
        return sns.pairplot(data=data.reset_index(), **kwargs), feature_names_df

    return sns.pairplot(data=data.reset_index(), **kwargs)


def _sanitize_palette(
    data: pd.DataFrame, palette: Optional[Union[str, Sequence[str]]], hue: str
) -> Optional[Union[str, Sequence[str]]]:
    if isinstance(palette, str):
        return palette
    if palette is not None and hue is not None:
        palette = palette[: len(data.reset_index()[hue].unique())]
    return palette


def _get_df_lineplot(data: pd.DataFrame, x: str, y: str, hue: str, order: Sequence[str]) -> pd.DataFrame:
    if "mean" in data.columns and "se" in data.columns:
        m_se = data
    else:
        group_cols = [x] if hue is None else [x, hue]
        m_se = data[[*group_cols, y]].groupby(group_cols).agg([np.mean, se])[y]
    if order is not None:
        m_se = m_se.reindex(order, level=0)
    return m_se


def _set_legend_errorbar(**kwargs):
    legend_fontsize = kwargs.get("legend_fontsize", "small")
    legend_loc = kwargs.get("legend_loc", "upper left")
    legend_title = kwargs.get("legend_title")
    ax = kwargs.get("ax")

    # get handles
    handles, labels = ax.get_legend_handles_labels()
    # remove the errorbars
    handles = [h[0] for h in handles]
    # use them in the legend
    ax.legend(handles, labels, numpoints=1, loc=legend_loc, title=legend_title, fontsize=legend_fontsize)


def _get_styles(
    data: pd.DataFrame,
    style: str,
    hue: str,
    marker: Optional[Union[str, Sequence[str]]] = None,
    linestyle: Optional[Union[str, Sequence[str]]] = None,
) -> tuple[Sequence[str], Sequence[str]]:
    num_cats = None
    if style is None:
        if all(v is not None for v in [hue, marker, linestyle]):
            num_cats = len(data[hue].unique())
            marker, linestyle = _get_marker_linestyle(marker, linestyle, num_cats)
        else:
            marker, linestyle = _get_marker_linestyle(marker, linestyle, 1)
    else:
        num_cats = len(data[style].unique())
        marker, linestyle = _get_marker_linestyle_style(marker, linestyle, num_cats)

    _get_styles_assert_styles(marker, linestyle, num_cats)

    return marker, linestyle


def _get_marker_linestyle(
    marker: Union[str, Sequence[str]], linestyle: Union[str, Sequence[str]], num_cats: int
) -> tuple[Sequence[str], Sequence[str]]:
    if isinstance(marker, str):
        marker = [marker] * num_cats
    if isinstance(linestyle, str):
        linestyle = [linestyle] * num_cats
    return marker, linestyle


def _get_marker_linestyle_style(
    marker: Union[str, Sequence[str]], linestyle: Union[str, Sequence[str]], num_cats: int
) -> tuple[Sequence[str], Sequence[str]]:
    if marker is None:
        marker = "o"
    if linestyle is None:
        linestyle = "-"
    return _get_marker_linestyle(marker, linestyle, num_cats)


def _get_styles_assert_styles(marker: Sequence[str], linestyle: Sequence[str], num_cats: int):
    if num_cats is not None:
        if len(marker) != num_cats:
            raise ValueError("If a list of 'marker' is provided it must match the number of 'style' categories!")
        if len(linestyle) != num_cats:
            raise ValueError("If a list of 'linestyle' is provided it must match the number of 'style' categories!")


def _add_stat_annot_multi_feature_boxplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    order: Sequence[str],
    hue: str,
    hue_order: Sequence[str],
    stats_kwargs: dict[str, Any],
    key: str,
    ax: plt.Axes,
    dict_box_pairs: Optional[dict[str, Sequence[str]]] = None,
    dict_pvals: Optional[dict[str, Sequence[float]]] = None,
):
    if len(stats_kwargs) > 0:
        stats_kwargs["comparisons_correction"] = stats_kwargs.get("comparisons_correction")
        stats_kwargs["test"] = stats_kwargs.get("test")

    if dict_box_pairs is not None:
        # filter box pairs by feature
        stats_kwargs["pairs"] = [dict_box_pairs[x] for x in dict_box_pairs if key in x]
        # flatten list
        stats_kwargs["pairs"] = [x for pairs in stats_kwargs["pairs"] for x in pairs]

    if dict_pvals is not None:
        # filter pvals by feature
        stats_kwargs["pvalues"] = [dict_pvals[x] for x in dict_pvals if key in x]
        # flatten list
        stats_kwargs["pvalues"] = [x for pairs in stats_kwargs["pvalues"] for x in pairs]
        stats_kwargs["perform_stat_test"] = False

    stats_kwargs["pvalue_thresholds"] = _PVALUE_THRESHOLDS

    if "pairs" in stats_kwargs and len(stats_kwargs["pairs"]) > 0:
        annotator = Annotator(
            data=data.reset_index(),
            ax=ax,
            x=x,
            y=y,
            order=order,
            hue=hue,
            hue_order=hue_order,
            **stats_kwargs,
        )
        annotator.configure(hide_non_significant=True, pvalue_thresholds=stats_kwargs.get("pvalue_thresholds"))
        annotator.set_pvalues(stats_kwargs.get("pvalues"))
        annotator.annotate()


def _feature_boxplot_add_legend(fig: plt.Figure, hue: str, handles: Sequence, labels: Sequence, **kwargs):
    legend_fontsize = kwargs.get("legend_fontsize")
    legend_loc = kwargs.get("legend_loc")
    legend_orientation = kwargs.get("legend_orientation")
    legend_title = kwargs.get("legend_title")
    rect = kwargs.get("rect")
    tight_layout = kwargs.get("tight_layout")

    if hue is not None:
        ncol = len(handles) if legend_orientation == "horizontal" else 1
        fig.legend(handles, labels, loc=legend_loc, fontsize=legend_fontsize, title=legend_title, ncol=ncol)
    if tight_layout:
        fig.tight_layout(pad=0.5, rect=rect)


def _multi_feature_boxplot_add_legend(fig: plt.Figure, hue: str, handles: Sequence, labels: Sequence, **kwargs):
    legend_fontsize = kwargs.get("legend_fontsize")
    legend_loc = kwargs.get("legend_loc")
    legend_orientation = kwargs.get("legend_orientation")
    legend_title = kwargs.get("legend_title")
    rect = kwargs.get("rect")

    if hue is not None:
        ncol = len(handles) if legend_orientation == "horizontal" else 1
        if handles is not None and len(handles) > 1:
            fig.legend(handles, labels, loc=legend_loc, ncol=ncol, fontsize=legend_fontsize, title=legend_title)
    if kwargs.pop("tight_layout", True):
        fig.tight_layout(pad=0.5, rect=rect)


def _style_xaxis_multi_feature_boxplot(
    ax: plt.Axes, xlabels: dict[str, str], xticklabels: dict[str, Union[str, Sequence[str]]], key: str
):
    if key in xticklabels:
        ax.set_xlabel(xlabels.get(key))
        xt = xticklabels[key]
        if isinstance(xt, str):
            xt = [xt]
        ax.set_xticks(np.arange(len(xt)))
        ax.set_xticklabels(xt)


def _plot_get_fig_ax(**kwargs):
    ax: plt.Axes = kwargs.get("ax")
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()
    return fig, ax


def _plot_get_fig_ax_list(
    features: dict[str, Union[str, Sequence[str]]], axs: list[plt.Axes], **kwargs
) -> tuple[plt.Figure, list[plt.Axes]]:
    figsize = kwargs.get("figsize")
    if axs is None:
        fig, axs = plt.subplots(figsize=figsize, ncols=len(features))
    else:
        fig = axs[0].get_figure()
    return fig, axs



"""Module providing various customized plotting functions."""

from biopsykit.plotting.plotting import (
    feature_boxplot,
    feature_pairplot,
    lineplot,
    multi_feature_boxplot,
    stacked_barchart,
)

__all__ = [
    "feature_boxplot",
    "feature_pairplot",
    "lineplot",
    "multi_feature_boxplot",
    "stacked_barchart",
]



"""Module for ECG segmentation."""

from biopsykit.signals.ecg.segmentation._base_segmentation import BaseHeartbeatSegmentation
from biopsykit.signals.ecg.segmentation._heartbeat_segmentation_neurokit import HeartbeatSegmentationNeurokit

__all__ = ["BaseHeartbeatSegmentation", "HeartbeatSegmentationNeurokit"]



"""Sleep/Wake detection using the *Cole/Kripke Algorithm*."""

import numpy as np
import pandas as pd

from biopsykit.sleep.sleep_wake_detection.algorithms._base import _SleepWakeBase
from biopsykit.sleep.sleep_wake_detection.utils import rescore
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import sanitize_input_1d
from biopsykit.utils.dtypes import SleepWakeDataFrame, _SleepWakeDataFrame


class ColeKripkeOld(_SleepWakeBase):
    """Class representing the *Cole/Kripke Algorithm* for sleep/wake detection based on activity counts."""

    def __init__(self, **kwargs):
        """Class representing the *Cole/Kripke Algorithm* for sleep/wake detection based on activity counts.

        The *Cole/Kripke Algorithm* runs sleep wake detection on epoch level activity data. Epochs are 1 minute long and
        activity is represented by an activity index which comes from Actigraph data or from raw acceleration data
        converted into activity index data.

        Parameters
        ----------
        scale_factor : float
            scale factor to use for the predictions (default corresponds to scale factor optimized for use with
            the activity index, if other activity measures are desired the scale factor can be modified or optimized.)
            The recommended range for the scale factor is between 0.1 and 0.25 depending on the sensitivity to activity
            desired, and possibly the population being observed.

        References
        ----------
        Cole, R. J., Kripke, D. F., Gruen, W., Mullaney, D. J., & Gillin, J. C. (1992). Automatic Sleep/Wake
        Identification From Wrist Activity. *Sleep*, 15(5), 461-469. https://doi.org/10.1093/sleep/15.5.461

        """
        self.scale_factor: float = kwargs.pop("scale_factor", None)
        """Scale factor to use for the predictions (default corresponds to scale factor optimized for use with the
        activity index, if other activity measures are desired the scale factor can be modified or optimized).
        The recommended range for the scale factor is between 0.1 and 0.25 depending on the sensitivity to activity
        desired, and possibly the population being observed.
        """

        if self.scale_factor is None:
            self.scale_factor = 0.193125
        super().__init__(**kwargs)

    def fit(self, data: arr_t, **kwargs) -> arr_t:  # noqa: ARG002
        """Fit sleep/wake detection algorithm to input data.

        .. note::
            Algorithms that do not have to (re)fit a ML model before sleep/wake prediction, such as rule-based
            algorithms, will internally bypass this method as the ``fit`` step is not needed.

        Parameters
        ----------
        data : array_like
            input data

        """
        return

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:  # noqa: ARG002
        """Apply sleep/wake prediction on activity index values.

        Parameters
        ----------
        data : array_like
            array with activity index values

        Returns
        -------
        :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
            dataframe with sleep/wake predictions

        """
        index = None
        if isinstance(data, pd.DataFrame):
            index = data.index
            data = sanitize_input_1d(data)

        # ensure numpy
        sf = np.array(self.scale_factor)
        kernel = sf * np.array([4.64, 6.87, 3.75, 5.07, 16.19, 5.84, 4.024, 0.00, 0.00])[::-1]
        scores = np.convolve(data, kernel, "same")

        scores[scores >= 1] = 99  # wake = 0
        scores[scores < 1] = 1  # sleep = 1
        scores[scores == 99] = 0  # wake = 0

        # rescore the original predictions
        scores = rescore(scores)

        scores = pd.DataFrame(scores, index=index, columns=["sleep_wake"])
        return _SleepWakeDataFrame(scores)



"""Module for ICG event extraction."""

from biopsykit.signals.icg.event_extraction._b_point_arbol2017 import (
    BPointExtractionArbol2017IsoelectricCrossings,
    BPointExtractionArbol2017SecondDerivative,
    BPointExtractionArbol2017ThirdDerivative,
)
from biopsykit.signals.icg.event_extraction._b_point_debski1993 import BPointExtractionDebski1993SecondDerivative
from biopsykit.signals.icg.event_extraction._b_point_drost2022 import BPointExtractionDrost2022
from biopsykit.signals.icg.event_extraction._b_point_forouzanfar2018 import BPointExtractionForouzanfar2018
from biopsykit.signals.icg.event_extraction._b_point_lozano2007 import (
    BPointExtractionLozano2007LinearRegression,
    BPointExtractionLozano2007QuadraticRegression,
)
from biopsykit.signals.icg.event_extraction._b_point_sherwood1990 import BPointExtractionSherwood1990
from biopsykit.signals.icg.event_extraction._b_point_stern1985 import BPointExtractionStern1985
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.signals.icg.event_extraction._base_c_point_extraction import BaseCPointExtraction
from biopsykit.signals.icg.event_extraction._c_point_scipy_findpeaks import CPointExtractionScipyFindPeaks

__all__ = [
    "BPointExtractionArbol2017IsoelectricCrossings",
    "BPointExtractionArbol2017SecondDerivative",
    "BPointExtractionArbol2017ThirdDerivative",
    "BPointExtractionDebski1993SecondDerivative",
    "BPointExtractionDrost2022",
    "BPointExtractionForouzanfar2018",
    "BPointExtractionLozano2007LinearRegression",
    "BPointExtractionLozano2007QuadraticRegression",
    "BPointExtractionSherwood1990",
    "BPointExtractionStern1985",
    "BaseBPointExtraction",
    "BaseCPointExtraction",
    "CPointExtractionScipyFindPeaks",
]



import warnings
from typing import Optional

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import (
    CPointDataFrame,
    HeartbeatSegmentationDataFrame,
    IcgRawDataFrame,
    is_b_point_dataframe,
    is_c_point_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_icg_raw_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError

__all__ = ["BPointExtractionStern1985"]


class BPointExtractionStern1985(BaseBPointExtraction, CanHandleMissingEventsMixin):
    """B-point extraction algorithm by Stern et al. (1985) [1]_.

    This algorithm extracts B-points based on the last local minimum of the dZ/dt curve before the C-point.

    References
    ----------
    .. [1] Stern, H. C., Wolf, G. K., & Belz, G. G. (1985). Comparative measurements of left ventricular ejection time
        by mechano-, echo- and electrical impedance cardiography. Arzneimittel-Forschung, 35(10), 1582-1586.

    """

    def __init__(self, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new ``BPointExtractionStern1985`` instance.

        Parameters
        ----------
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle failing event extraction. Can be one of:
                * "warn": issue a warning and set the event to NaN
                * "raise": raise an ``EventExtractionError``
                * "ignore": ignore the error and continue with the next event
            Default: "warn"

        """
        super().__init__(handle_missing_events=handle_missing_events)

    # @make_action_safe
    def extract(
        self,
        *,
        icg: IcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: Optional[float],  # noqa: ARG002
    ):
        """Extract B-points from given ICG derivative signal.

        This algorithm extracts B-points based on the last local minimum of the dZ/dt curve before the C-point.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        icg : :class:`~pandas.DataFrame`
            ICG derivative signal
        heartbeats : :class:`~pandas.DataFrame`
            Segmented heartbeats. Each row contains start, end, and R-peak location (in samples
            from beginning of signal) of that heartbeat, index functions as id of heartbeat
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        self._check_valid_missing_handling()
        is_icg_raw_dataframe(icg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        is_c_point_dataframe(c_points)
        # sanitize input
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        # Create the b_point Dataframe. Use the heartbeats id as index
        b_points = pd.DataFrame(index=heartbeats.index, columns=["b_point_sample", "nan_reason"])

        # get the c_point locations from the c_points dataframe and search for entries containing NaN
        c_points = c_points["c_point_sample"]
        check_c_points = pd.isna(c_points)

        # Compute the second derivative of the ICG-signal
        icg_2nd_der = np.gradient(icg)
        icg_der_zero_crossings = np.where(np.diff(np.signbit(icg_2nd_der)))[0]

        # go through each heartbeat independently and search for the local minima
        for idx, data in heartbeats.iterrows():
            heartbeat_start = data["start_sample"]

            # check c_point is NaN. If this is the case, set the b_point to NaN and continue with the next iteration
            if check_c_points[idx]:
                b_points.loc[idx, "b_point_sample"] = np.NaN
                b_points.loc[idx, "nan_reason"] = "c_point_nan"
                missing_str = f"The c_point is NaN at position {idx}! B-Point was set to NaN."
                if self.handle_missing_events == "warn":
                    warnings.warn(missing_str)
                elif self.handle_missing_events == "raise":
                    raise EventExtractionError(missing_str)
                continue

            # check if there are zero crossings in the interval between start of the heartbeat and the C-point
            # we subtract 1 to avoid the C-point itself since the zero crossing should be *before* the C-point and the
            # zero crossings are computed in a way that the sample before the zero crossing is returned
            zero_crossings_heartbeat = icg_der_zero_crossings[
                (icg_der_zero_crossings >= heartbeat_start) & (icg_der_zero_crossings < (c_points[idx] - 1))
            ]

            # if there are no zero crossings in the interval, set B-point to NaN
            if len(zero_crossings_heartbeat) == 0:
                b_points.loc[idx, "b_point_sample"] = np.NaN
                b_points.loc[idx, "nan_reason"] = "no_local_minimum"
                continue
            # get the closest zero crossing *before* the C-point
            zero_crossings_diff = zero_crossings_heartbeat - c_points[idx]
            zero_crossings_diff = zero_crossings_diff[zero_crossings_diff < 0]
            zero_crossing_idx = np.argmax(zero_crossings_diff)

            b_point = zero_crossings_heartbeat[zero_crossing_idx]

            b_points.loc[idx, "b_point_sample"] = b_point

        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        is_b_point_dataframe(b_points)

        self.points_ = b_points

        return self

    def _b_point_core_extraction(
        self,
        icg_2nd_der: pd.Series,
        r_peak: pd.Series,
        c_point: pd.Series,
    ):
        # set the borders of the interval between the R-Peak and the C-Point
        start_r_c = r_peak
        end_r_c = c_point

        # Select the specific interval in the second derivative of the ICG-signal
        icg_search_window = icg_2nd_der[start_r_c : (end_r_c + 1)]

        # Compute the local minima in this interval
        # icg_min = argrelmin(icg_search_window)
        icg_min = find_peaks(-icg_search_window)[0]
        # print(icg_min)

        # Compute the distance between the C-point and the minima of the interval and select the entry with
        # the minimal distance as B-point
        if len(icg_min) >= 1:
            distance = end_r_c - icg_min
            b_point_idx = distance.argmin()
            b_point = icg_min[b_point_idx]
            # Compute the absolute sample position of the local B-point
            b_point = b_point + start_r_c
        else:
            # If there is no minima set the B-Point to NaN
            b_point = np.NaN

        return b_point



import warnings
from typing import Optional

import numpy as np
import pandas as pd
from scipy.signal import argrelmin
from tpcp import Parameter

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import (
    CPointDataFrame,
    HeartbeatSegmentationDataFrame,
    IcgRawDataFrame,
    is_b_point_dataframe,
    is_c_point_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_icg_raw_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError

__all__ = ["BPointExtractionForouzanfar2018"]


class BPointExtractionForouzanfar2018(BaseBPointExtraction, CanHandleMissingEventsMixin):
    """B-point extraction algorithm by Forouzanfar et al. (2018) [1]_.

    This algorithm extracts B-points based on the detection of the most prominent monotonic increasing segment
    between the A-Point (the local minimum within one third of the beat-to-beat interval prior to the C-Point)
    and the C-Point. The B-Point is then detected as the last zero crossing or local maximum of the third derivative
    of the ICG signal within the segment.

    References
    ----------
    .. [1] Forouzanfar, M., Baker, F. C., De Zambotti, M., McCall, C., Giovangrandi, L., & Kovacs, G. T. A. (2018).
        Toward a better noninvasive assessment of preejection period: A novel automatic algorithm for B-point detection
        and correction on thoracic impedance cardiogram. Psychophysiology, 55(8), e13072.
        https://doi.org/10.1111/psyp.13072

    """

    # input parameters
    scaling_factor: Parameter[float]
    correct_outliers: Parameter[bool]

    def __init__(
        self,
        scaling_factor: float = 2000,
        correct_outliers: Optional[bool] = False,
        handle_missing_events: HANDLE_MISSING_EVENTS = "warn",
    ):
        """Initialize new ``BPointExtractionForouzanfar2018`` instance.

        .. note:: In the original paper, the authors report the *sampling frequency* of the ICG signal as the scaling
        factor. Since this would change the algorithm behavior depending on the sampling rate of the ICG signal,
        this implementation introduces a scaling factor that can be set by the user. By default, the scaling factor is
        set to 2000 (corresponding to a sampling rate of the original data of 2000 Hz) instead of using the
        sampling rate of the ICG signal.

        Parameters
        ----------
        scaling_factor : float, optional
            Scaling factor for the B-point extraction algorithm. Default: 2000
        correct_outliers : bool, optional
            True to correct outliers, False to set B-Point to NaN if no monotonic segment is found. Default: False
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle failing event extraction. Can be one of:
                * "warn": issue a warning and set the event to NaN
                * "raise": raise an ``EventExtractionError``
                * "ignore": ignore the error and continue with the next event
            Default: "warn"

        """
        super().__init__(handle_missing_events=handle_missing_events)
        self.scaling_factor = scaling_factor
        self.correct_outliers = correct_outliers

    # @make_action_safe
    def extract(  # noqa: PLR0915
        self,
        *,
        icg: IcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: Optional[float],  # noqa: ARG002
    ):
        """Extract B-points from given ICG derivative signal.

        The algorithm extracts B-points based on the detection of the most prominent monotonic increasing segment
        between the A-Point (the local minimum within one third of the beat-to-beat interval prior to the C-Point)
        and the C-Point. The B-Point is then detected as the last zero crossing or local maximum of the third derivative
        of the ICG signal within the segment.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        icg : :class:`~pandas.DataFrame`
            ICG derivative signal
        heartbeats : :class:`~pandas.DataFrame`
            Segmented heartbeats. Each row contains start, end, and R-peak location (in samples
            from beginning of signal) of that heartbeat, index functions as id of heartbeat
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz

        Returns
        -------
        self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        # sanitize input signal
        self._check_valid_missing_handling()
        is_icg_raw_dataframe(icg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        is_c_point_dataframe(c_points)
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        # Create the B-Point/A-Point Dataframes with the index of the heartbeat_list
        b_points = pd.DataFrame(index=heartbeats.index, columns=["b_point_sample", "nan_reason"])

        # check whether the c_points contain NaN
        check_c_points = pd.isna(c_points["c_point_sample"]).to_numpy()

        # Calculate the second- and third-derivative of the ICG-signal
        second_der = np.gradient(icg)
        third_der = np.gradient(second_der)

        for idx, data in heartbeats.iloc[1:].iterrows():
            # check if the current or the previous C-Point contain NaN . If this is the case, set the b_point to NaN
            if check_c_points[idx] or check_c_points[idx - 1]:
                b_points.loc[idx, "b_point_sample"] = np.nan
                b_points.loc[idx, "nan_reason"] = "c_point_nan"
                continue

            # Detect the main peak in the dZ/dt signal (C-Point)
            c_point = c_points.loc[idx, "c_point_sample"]

            # Compute the beat to beat interval
            beat_to_beat = c_points.loc[idx, "c_point_sample"] - c_points.loc[idx - 1, "c_point_sample"]

            # Compute the search interval for the A-Point
            search_interval = int(beat_to_beat / 3)

            # Detect the local minimum (A-Point) within one third of the beat to beat interval prior to the C-Point
            a_point = self._get_a_point(icg, search_interval, c_point) + (c_point - search_interval)

            # Select the signal_segment between the A-Point and the C-Point
            signal_clean_segment = icg.iloc[a_point : c_point + 1]

            # Define the C-Point amplitude which is used as a constraint for monotonic segment detection
            c_amplitude = icg.iloc[c_point]

            # Step 4.1: Get the most prominent monotonic increasing segment between the A-Point and the C-Point
            start_sample, end_sample = (
                self._get_most_prominent_monotonic_increasing_segment(signal_clean_segment, c_amplitude) + a_point
            )

            if (start_sample == a_point) & (end_sample == a_point):
                if self.correct_outliers:
                    b_points.loc[idx, "b_point_sample"] = data["r_peak_sample"]
                else:
                    b_points.loc[idx, "b_point_sample"] = np.nan
                    b_points.loc[idx, "nan_reason"] = "no_monotonic_segment"
                continue

            # Get the first third of the monotonic increasing segment
            start = start_sample
            end = end_sample - int((2 / 3) * (end_sample - start_sample))

            # 2nd derivative of the segment
            monotonic_segment_2nd_der = pd.DataFrame(second_der[start:end], columns=["2nd_der"])
            # 3rd derivative of the segment
            monotonic_segment_3rd_der = pd.DataFrame(third_der[start:end], columns=["3rd_der"])

            # Calculate the amplitude difference between the C-Point and the A-Point
            height = icg.iloc[c_point] - icg.iloc[a_point]

            # Compute the significant zero_crossings
            significant_zero_crossings = self._get_zero_crossings_3rd_derivative(
                monotonic_segment_3rd_der, monotonic_segment_2nd_der, height
            )

            # Compute the significant local maxima of the 3rd derivative of the most prominent monotonic segment
            significant_local_maxima = self._get_local_maxima_3rd_derivative(monotonic_segment_3rd_der, height)

            # Label the last zero crossing/ local maximum as the B-Point
            # If there are no zero crossings or local maxima use the first Point of the segment as B-Point
            significant_features = pd.concat([significant_zero_crossings, significant_local_maxima], axis=0) + start
            if len(significant_features) == 0:
                b_point = start
            else:
                b_point = significant_features.iloc[np.argmin(c_point - significant_features)].iloc[0]

            b_points.loc[idx, "b_point_sample"] = b_point

        # interpolate the first B-Point with the second B-Point since it is not possible to detect a B-Point
        # for the first heartbeat
        b_points.loc[b_points.index[0], "b_point_sample"] = (
            b_points.loc[b_points.index[1], "b_point_sample"] - b_points["b_point_sample"].diff().loc[b_points.index[2]]
        )

        idx_nan = b_points["b_point_sample"].isna()
        if idx_nan.sum() > 0:
            idx_nan = list(b_points.index[idx_nan])
            missing_str = (
                f"Either 'r_peak' or 'c_point' contains NaN at positions {idx_nan}! The B-points were set to NaN."
            )
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)

        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        is_b_point_dataframe(b_points)

        self.points_ = b_points
        return self

    @staticmethod
    def _get_a_point(icg: pd.Series, search_interval: int, c_point: int):
        signal_interval = icg.iloc[(c_point - search_interval) : c_point]
        signal_minima = argrelmin(signal_interval.values, mode="wrap")

        a_point_candidates = signal_interval.iloc[signal_minima[0]]
        if len(a_point_candidates) == 0:
            # no local minima found => return the argmin of the interval
            return np.argmin(signal_interval)

        a_point_idx = np.argmin(a_point_candidates)
        a_point = signal_minima[0][a_point_idx]
        return a_point

    @staticmethod
    def _get_most_prominent_monotonic_increasing_segment(icg_segment: pd.Series, height: int):
        icg_segment = icg_segment.copy()
        icg_2nd_der_segment = np.gradient(icg_segment)
        icg_segment.index = np.arange(0, len(icg_segment))
        monotony_df = pd.DataFrame(icg_segment.values, columns=["icg"])
        monotony_df = monotony_df.assign(**{"2nd_der": icg_2nd_der_segment, "borders": pd.NA})

        # A-Point is a possible start of the monotonic segment
        monotony_df.loc[monotony_df.index[0], "borders"] = "start_increase"
        # C-Point is a possible end of the monotonic segment
        monotony_df.loc[monotony_df.index[-1], "borders"] = "end_increase"

        neg_pos_change_idx = np.where(np.diff(np.sign(monotony_df["2nd_der"])) > 0)[0]
        monotony_df.loc[monotony_df.index[neg_pos_change_idx], "borders"] = "start_increase"

        pos_neg_change_idx = np.where(np.diff(np.sign(monotony_df["2nd_der"])) < 0)[0]
        monotony_df.loc[monotony_df.index[pos_neg_change_idx], "borders"] = "end_increase"

        # drop all samples that are no possible start-/ end-points
        monotony_df = monotony_df.drop(monotony_df[pd.isna(monotony_df["borders"])].index)
        monotony_df = monotony_df.reset_index()

        # Drop start- and corresponding end-point if their start value is higher than 1/2 of H
        start_index_drop_rule_a = monotony_df[
            (monotony_df["borders"] == "start_increase") & (monotony_df["icg"] > height / 2)
        ].index
        start_index_drop_rule_a = start_index_drop_rule_a.union(start_index_drop_rule_a + 1)
        monotony_df = monotony_df.drop(index=start_index_drop_rule_a)

        # Drop start- and corresponding end-point if their end values does not reach at least 2/3 of H
        end_index_drop_rule_b = monotony_df[
            (monotony_df["borders"] == "end_increase") & (monotony_df["icg"] < 2 * height / 3)
        ].index

        end_index_drop_rule_b = end_index_drop_rule_b.union(end_index_drop_rule_b - 1)
        monotony_df = monotony_df.drop(index=monotony_df.iloc[end_index_drop_rule_b].index)

        # Select the monotonic segment with the highest amplitude difference
        start_sample = 0
        end_sample = 0
        if len(monotony_df) > 2:
            idx = np.argmax(monotony_df["icg"].diff())
            start_sample = monotony_df["index"].iloc[idx - 1]
            end_sample = monotony_df["index"].iloc[idx]
        elif len(monotony_df) != 0:
            start_sample = monotony_df["index"].iloc[0]
            end_sample = monotony_df["index"].iloc[-1]

        return start_sample, end_sample  # That are not absolute positions yet

    def _get_zero_crossings_3rd_derivative(
        self, monotonic_segment_3rd_der: pd.DataFrame, monotonic_segment_2nd_der: pd.DataFrame, height: int
    ):
        constraint = float(10 * height / self.scaling_factor)

        zero_crossings = np.where(np.diff(np.signbit(monotonic_segment_3rd_der["3rd_der"])))[0]
        zero_crossings = pd.DataFrame(zero_crossings, columns=["sample_position"])

        # Discard zero_crossings with negative to positive sign change
        significant_crossings = zero_crossings.drop(
            zero_crossings[monotonic_segment_2nd_der.iloc[zero_crossings["sample_position"]].to_numpy() < 0].index,
            axis=0,
        )

        # Discard zero crossings with slope higher than 10*H/scaling_factor
        significant_crossings = significant_crossings.drop(
            significant_crossings[
                monotonic_segment_2nd_der.iloc[significant_crossings["sample_position"]].to_numpy() >= constraint
            ].index,
            axis=0,
        )

        if isinstance(zero_crossings, type(None)) or len(zero_crossings) == 0:
            return pd.DataFrame([0], columns=["sample_position"])
        return significant_crossings

    def _get_local_maxima_3rd_derivative(self, monotonic_segment_3rd_der: pd.DataFrame, height: int):
        constraint = float(4 * height / self.scaling_factor)
        # compute gradient
        monotonic_segment_3rd_der_gradient = np.gradient(monotonic_segment_3rd_der.squeeze())

        # find zero-crossings of the gradient to determine local maxima
        local_maxima = np.where(np.diff(np.sign(monotonic_segment_3rd_der_gradient)) < 0)[0]
        # add 1 to the index to get the correct position
        local_maxima += 1
        local_maxima = pd.DataFrame(local_maxima, columns=["sample_position"])

        significant_maxima = local_maxima.drop(
            local_maxima[
                monotonic_segment_3rd_der["3rd_der"].iloc[local_maxima["sample_position"]].to_numpy() < constraint
            ].index,
            axis=0,
        )

        if isinstance(significant_maxima, type(None)):
            return pd.DataFrame([0], columns=["sample_position"])
        if len(significant_maxima) == 0:
            return pd.DataFrame([0], columns=["sample_position"])
        return significant_maxima



from typing import Optional

from biopsykit.signals.icg.outlier_correction import BaseOutlierCorrection

__all__ = ["OutlierCorrectionDummy"]

from biopsykit.utils.dtypes import BPointDataFrame, CPointDataFrame, is_b_point_dataframe


class OutlierCorrectionDummy(BaseOutlierCorrection):
    """B-point outlier correction algorithm that does nothing and passes through the input data unchanged.

    This class is used as a placeholder for the outlier correction in the ICG pipeline.

    """

    def correct_outlier(
        self,
        *,
        b_points: BPointDataFrame,
        c_points: Optional[CPointDataFrame],  # noqa: ARG002
        sampling_rate_hz: Optional[float],  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ):
        """Perform outlier correction.

        This method does nothing and passes through the input data unchanged.

        The results of the outlier correction are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        b_points : :class:`~pandas.DataFrame`
            Extracted B-points. Each row contains the B-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. B-point locations can be NaN if no B-points were detected
            for certain heartbeats.
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats.
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz
        kwargs: dict
            Additional keyword arguments. Not used.

        Returns
        -------
        self

        """
        is_b_point_dataframe(b_points)
        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})

        self.points_ = b_points
        return self



from typing import Optional

import pandas as pd
from tpcp import Parameter

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.ecg.event_extraction._base_ecg_extraction import BaseEcgExtraction
from biopsykit.utils.dtypes import (
    EcgRawDataFrame,
    HeartbeatSegmentationDataFrame,
    is_ecg_raw_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_q_peak_dataframe,
)


class QPeakExtractionVanLien2013(BaseEcgExtraction, CanHandleMissingEventsMixin):
    """Algorithm to extract Q-peaks based on the detection of the R-peak, as suggested by Van Lien et al. (2013) [1]_.

    The Q-peak is estimated by subtracting a fixed time interval from the R-peak location. The fixed time
    interval is defined by the parameter ``time_interval``.


    References
    ----------
    .. [1] Van Lien, R., Schutte, N. M., Meijer, J. H., & De Geus, E. J. C. (2013). Estimated preejection period (PEP)
        based on the detection of the R-peak and dZ/dt-min peaks does not adequately reflect the actual PEP across a
        wide range of laboratory and ambulatory conditions. International Journal of Psychophysiology, 87(1), 60-69.
        https://doi.org/10.1016/j.ijpsycho.2012.11.001

    """

    # parameters
    time_interval_ms: Parameter[int]

    def __init__(self, time_interval_ms: int = 40, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new ``QPeakExtractionVanLien2013`` algorithm instance.

        Parameters
        ----------
        time_interval_ms : int, optional
            Specify the constant time interval in milliseconds which will be subtracted from the R-peak for
            Q-peak estimation. Default: 40 ms
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle missing data in the input dataframes. Default: "warn"
        """
        super().__init__(handle_missing_events=handle_missing_events)
        self.time_interval_ms = time_interval_ms

    # @make_action_safe
    def extract(
        self,
        *,
        ecg: Optional[EcgRawDataFrame],
        heartbeats: HeartbeatSegmentationDataFrame,
        sampling_rate_hz: float,
    ):
        """Extract Q-peaks from given ECG signal.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        ecg: :class:`~pandas.DataFrame`
            ECG signal. Not used in this function since Q-peak is estimated from the R-peaks in the
            ``heartbeats`` DataFrame.
        heartbeats: :class:`~pandas.DataFrame`
            DataFrame containing one row per segmented heartbeat, each row contains start, end, and R-peak
            location (in samples from beginning of signal) of that heartbeat, index functions as id of heartbeat
        sampling_rate_hz: int
            Sampling rate of ECG signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        is_ecg_raw_dataframe(ecg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        # TODO handle missing?
        # convert the fixed time_interval from milliseconds into samples
        time_interval_in_samples = int((self.time_interval_ms / 1000) * sampling_rate_hz)

        # get the r_peaks from the heartbeats Dataframe
        r_peaks = heartbeats[["r_peak_sample"]]

        # subtract the fixed time_interval from the r_peak samples to estimate the q_peaks
        q_peaks = r_peaks - time_interval_in_samples

        q_peaks.columns = ["q_peak_sample"]
        q_peaks = q_peaks.assign(nan_reason=pd.NA)

        q_peaks = q_peaks.astype({"q_peak_sample": "Int64", "nan_reason": "object"})
        is_q_peak_dataframe(q_peaks)

        self.points_ = q_peaks
        return self



import pandas as pd
from scipy import signal
from tpcp import Algorithm, Parameter

from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import is_icg_raw_dataframe


class IcgPreprocessingBandpass(Algorithm):
    _action_methods = "clean"

    method: Parameter[str]

    icg_clean_: pd.DataFrame

    def __init__(self, method: str = "butterworth"):
        self.method = method

    def clean(self, *, icg: pd.DataFrame, sampling_rate_hz: int):
        """Clean ICG signal using a band-pass filter.

        Parameters
        ----------
        icg : :class:`~pandas.DataFrame`
            pandas DataFrame containing the raw ICG signal
        sampling_rate_hz : int
            Sampling rate of the ICG signal in Hz

        Returns
        -------
        self

        """
        is_icg_raw_dataframe(icg)
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        if icg.empty:
            self.icg_clean_ = pd.DataFrame(index=icg.index, columns=["icg_der"])
            return self

        if self.method not in ["butterworth", "elliptic", "savgol"]:
            raise ValueError("Filter type can only be 'butterworth', 'elliptic', or 'savgol'")

        if self.method == "butterworth":
            sos = signal.butter(N=4, Wn=[0.5, 25], btype="bandpass", output="sos", fs=sampling_rate_hz)
            clean_signal = signal.sosfiltfilt(sos, icg)
        elif self.method == "elliptic":
            rp = 1.0
            rs = 80.0
            sos = signal.ellip(
                N=2, rp=rp, rs=rs, Wn=[0.75, 40], btype="bandpass", output="sos", fs=sampling_rate_hz, analog=False
            )
            clean_signal = signal.sosfiltfilt(sos, icg)
        elif self.method == "savgol":  # Savitzky-Golay filter (for high frequency noise?!)
            clean_signal = signal.savgol_filter(icg, window_length=61, polyorder=3)
        else:
            raise ValueError("This should not happen")

        self.icg_clean_ = pd.DataFrame(clean_signal, index=icg.index, columns=["icg_der"])
        return self



from typing import Literal, get_args

import pandas as pd
from tpcp import Algorithm

HANDLE_MISSING_EVENTS = Literal["raise", "warn", "ignore"]


class CanHandleMissingEventsMixin(Algorithm):
    def __init__(self, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Mixin class to handle missing events in the input dataframes.

        Parameters
        ----------
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle missing data in the input dataframes. If "warn", a warning is raised if missing data is found.
            If "raise", an exception is raised if missing data is found. If "ignore", missing data is ignored.
            Default: "warn"

        """
        self.handle_missing_events = handle_missing_events

    def _check_valid_missing_handling(self):
        if self.handle_missing_events not in get_args(HANDLE_MISSING_EVENTS):
            raise ValueError(
                f"Invalid value '{self.handle_missing_events}' for 'handle_missing_events'. "
                f"Must be one of {get_args(HANDLE_MISSING_EVENTS)}."
            )


class BaseExtraction(Algorithm):
    """Base class which defines the interface for all fiducial point extraction algorithms.

    Results:
        points_ : saves positions of extracted points in pd.DataFrame
    """

    _action_methods = "extract"

    # results
    points_: pd.DataFrame



"""Module for systematically evaluating different combinations of sklearn pipelines."""

import functools
import pickle
import re
from collections.abc import Sequence
from copy import deepcopy
from inspect import getmembers, signature
from itertools import product
from pathlib import Path
from shutil import rmtree
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
import sklearn.metrics
from joblib import Memory
from numpy.random import RandomState
from sklearn.base import BaseEstimator, clone
from sklearn.model_selection import BaseCrossValidator
from sklearn.pipeline import Pipeline
from tqdm.auto import tqdm
from typing_extensions import Self

from biopsykit.classification.model_selection import nested_cv_param_search
from biopsykit.classification.utils import _PipelineWrapper, merge_nested_dicts
from biopsykit.utils._datatype_validation_helper import _assert_file_extension
from biopsykit.utils._types_internal import path_t, str_t

__all__ = ["SklearnPipelinePermuter"]

pipeline_step_map = {
    "pipeline_scaler": "Scaler",
    "pipeline_reduce_dim": r"\makecell[lc]{Feature\\ Selection}",
    "pipeline_clf": "Classifier",
}

metric_map = {
    "accuracy": r"\makecell{Accuracy [\%]}",
    "f1": r"\makecell{F1-score [\%]}",
    "precision": r"\makecell{Precision [\%]}",
    "recall": r"\makecell{Recall [\%]}",
    "auc": r"\makecell{AUC [\%]}",
    "sensitivity": r"\makecell{Sensitivity [\%]}",
    "specificity": r"\makecell{Specificity [\%]}",
}

clf_map = {
    "MinMaxScaler": "Min-Max",
    "StandardScaler": "Standard",
    "SelectKBest": "SkB",
    "RFE": "RFE",
    "SelectFromModel": "SFM",
    "GaussianNB": "NB",
    "KNeighborsClassifier": "kNN",
    "DecisionTreeClassifier": "DT",
    "SVC": "SVM",
    "RandomForestClassifier": "RF",
    "MLPClassifier": "MLP",
    "AdaBoostClassifier": "Ada",
}


class SklearnPipelinePermuter:
    """Class for systematically evaluating different sklearn pipeline combinations."""

    def __init__(
        self,
        model_dict: Optional[dict[str, dict[str, BaseEstimator]]] = None,
        param_dict: Optional[dict[str, Optional[Union[Sequence[dict[str, Any]], dict[str, Any]]]]] = None,
        hyper_search_dict: Optional[dict[str, dict[str, Any]]] = None,
        random_state: Optional[int] = None,
    ):
        """Class for systematically evaluating different sklearn pipeline combinations.

        This class can be used to, for instance, evaluate combinations of different feature selection methods
        (e.g., :class:`~sklearn.feature_selection.SelectKBest`,
        :class:`~sklearn.feature_selection.SequentialFeatureSelector`) with different estimators
        (e.g., :class:`~sklearn.svm.SVC`, :class:`~sklearn.tree.DecisionTreeClassifier`), any much more.

        For all combinations, hyperparameter search (e.g., using grid-search or randomized-search) can be performed by
        passing one joint parameter grid (see Examples).

        Parameters
        ----------
        model_dict : dict
            Dictionary specifying the different transformers and estimators to evaluate.
            Each pipeline step corresponds to one dictionary entry and has the name of the pipeline step (str) as key.
            The values are again dictionaries with the transformer/estimator names as keys and instances of the
            transformers/estimators as values
        param_dict : dict
            Nested dictionary specifying the parameter settings to try per transformer/estimator. The dictionary has
            the transformer/estimator names (str) as keys and parameter dictionaries as values. Each parameter
            dictionary has parameters names (str) as keys and lists of parameter settings to try as values, or a list
            of such dictionaries, in which case the grids spanned by each dictionary in the list are explored.
            This enables searching over any sequence of parameter settings.
        hyper_search_dict : dict, optional
            Nested dictionary specifying the method for hyperparameter search (e.g., whether to use "grid" for
            grid-search or "random" for randomized-search) for each estimator. By default, "grid-search" is used
            for each estimator unless individually specified otherwise.
        random_state : int, optional
            Controls the random seed passed to each estimator and each splitter. By default, no random seed is passed.
            Set this to an integer for reproducible results across multiple program calls.

        Examples
        --------
        >>> from sklearn import datasets
        >>> from sklearn.preprocessing import StandardScaler, MinMaxScaler
        >>> from sklearn.feature_selection import SelectKBest, RFE
        >>> from sklearn.neighbors import KNeighborsClassifier
        >>> from sklearn.svm import SVC
        >>> from sklearn.tree import DecisionTreeClassifier
        >>> from sklearn.ensemble import AdaBoostClassifier
        >>> from sklearn.model_selection import KFold
        >>>
        >>> from biopsykit.classification.model_selection import SklearnPipelinePermuter
        >>>
        >>> breast_cancer = datasets.load_breast_cancer()
        >>> X = breast_cancer.data
        >>> y = breast_cancer.target
        >>>
        >>> model_dict = {
        >>>    "scaler": {
        >>>         "StandardScaler": StandardScaler(),
        >>>         "MinMaxScaler": MinMaxScaler(),
        >>>     },
        >>>     "reduce_dim": {
        >>>         "SelectKBest": SelectKBest(),
        >>>         "RFE": RFE(SVC(kernel="linear", C=1))
        >>>     },
        >>>     "clf" : {
        >>>         "KNeighborsClassifier": KNeighborsClassifier(),
        >>>         "DecisionTreeClassifier": DecisionTreeClassifier(),
        >>>         "SVC": SVC(),
        >>>         "AdaBoostClassifier": AdaBoostClassifier(),
        >>>     }
        >>> }
        >>>
        >>> param_dict = {
        >>>     "StandardScaler": None,
        >>>     "MinMaxScaler": None,
        >>>     "SelectKBest": { "k": [2, 4, 6, 8, "all"] },
        >>>     "RFE": { "n_features_to_select": [2, 4, 6, 8, None] },
        >>>     "KNeighborsClassifier": { "n_neighbors": [2, 4, 6, 8], "weights": ["uniform", "distance"] },
        >>>     "DecisionTreeClassifier": {"criterion": ['gini', 'entropy'], "max_depth": [2, 4, 6, 8, 10] },
        >>>     "AdaBoostClassifier": {
        >>>         "base_estimator": [DecisionTreeClassifier(max_depth=1), DecisionTreeClassifier(max_depth=2)],
        >>>         "n_estimators": np.arange(20, 210, 10),
        >>>         "learning_rate": np.arange(0.6, 1.1, 0.1)
        >>>     },
        >>>     "SVC": [
        >>>         {
        >>>             "kernel": ["linear"],
        >>>             "C": np.logspace(start=-3, stop=3, num=7)
        >>>         },
        >>>         {
        >>>             "kernel": ["rbf"],
        >>>             "C": np.logspace(start=-3, stop=3, num=7),
        >>>             "gamma": np.logspace(start=-3, stop=3, num=7)
        >>>         }
        >>>     ]
        >>> }
        >>>
        >>> # AdaBoost hyperparameters should be optimized using randomized-search, all others using grid-search
        >>> hyper_search_dict = {
        >>>     "AdaBoostClassifier": {"search_method": "random", "n_iter": 30}
        >>> }
        >>>
        >>> pipeline_permuter = SklearnPipelinePermuter(model_dict, param_dict, hyper_search_dict)
        >>> pipeline_permuter.fit(X, y, outer_cv=KFold(), inner_cv=KFold())

        """
        self.models: dict[str, dict[str, BaseEstimator]] = {}
        """Dictionary with pipeline steps and the different transformers/estimators per step."""

        self.params: dict[str, Optional[Union[Sequence[dict[str, Any]], dict[str, Any]]]] = {}
        """Dictionary with parameter sets to test for the different transformers/estimators per pipeline step."""

        self.model_combinations: Sequence[tuple[tuple[str, str], ...]] = []
        """List of model combinations, i.e. permutations of the different transformers/estimators for
        each pipeline step."""

        self.hyper_search_dict: dict[str, dict[str, Any]] = {}
        """Dictionary specifying the selected hyperparameter search method for each estimator."""

        self.param_searches: dict[tuple[str, str], dict[str, Any]] = {}
        """Dictionary with parameter search results for each pipeline step combination."""

        self.results: Optional[pd.DataFrame] = None
        """Dataframe with parameter search results of each pipeline step combination."""

        self.scoring: str_t = ""
        """Scoring used as metric for optimization during hyperparameter search."""

        self.refit: str = ""

        self.random_state: Optional[RandomState] = None

        self._results_set: bool = False

        if model_dict is None and param_dict is None:
            # create empty instance
            return

        self.random_state = RandomState(random_state)

        self._set_permuter_params(model_dict, param_dict, hyper_search_dict)

    def _set_permuter_params(self, model_dict, param_dict, hyper_search_dict):
        self._check_missing_params(model_dict, param_dict)

        if hyper_search_dict is None:
            hyper_search_dict = {}
        self.hyper_search_dict = hyper_search_dict.copy()

        clf_list = model_dict[list(model_dict.keys())[-1]]
        for clf in clf_list:
            # fill the dict with the default search method (grid-search) for the classifiers that are not
            # specified explicitly
            self.hyper_search_dict.setdefault(clf, {"search_method": "grid"})

        model_combinations = list(product(*[[(step, k) for k in list(model_dict[step].keys())] for step in model_dict]))

        # assert that all entries of the param dict are lists for uniform handling
        for k, v in param_dict.items():
            if isinstance(v, dict):
                param_dict[k] = [v]

        model_dict = deepcopy(model_dict)
        self.models = self._initialize_models(model_dict)
        self.params = param_dict
        self.model_combinations = model_combinations

    def _initialize_models(self, model_dict: dict[str, dict[str, Any]]) -> dict[str, dict[str, Any]]:
        if self.random_state is None:
            return model_dict
        for _k, v in model_dict.items():
            # add fixed random state to each estimator if it has a random_state parameter
            for estimator in v.values():
                if hasattr(estimator, "random_state"):
                    estimator.random_state = self.random_state
        return model_dict

    @property
    def results(self):
        """Parameter search results of each pipeline step combination.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Dataframe with parameter search results of each pipeline step combination

        """
        if self._results is None:
            self._results = self.pipeline_score_results()
        return self._results

    @results.setter
    def results(self, results):
        if results is None:
            self._results_set = False
        else:
            self._results_set = True
        self._results = results

    @classmethod
    def from_csv(cls, file_path: path_t, num_pipeline_steps: Optional[int] = 3) -> Self:
        """Create a new ``SklearnPipelinePermuter`` instance from a csv file with results from parameter search.

        Parameters
        ----------
        file_path : :class:`pathlib.Path` or str
            path to csv file
        num_pipeline_steps : int
            integer specifying the number of steps in the pipeline. Used to infer pipeline steps from the
            :class:`~pandas.MultiIndex` in the dataframe. For instance, for a pipeline consisting of the steps
            "scaler", "reduce_dim", and "clf" pass "3" as ``num_pipeline_steps``

        Returns
        -------
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
            ``SklearnPipelinePermuter`` instance with results from csv file

        """
        # assert pathlib
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".csv")
        score_summary = pd.read_csv(file_path)
        score_summary = score_summary.set_index(list(score_summary.columns)[: num_pipeline_steps + 2])
        pipeline_permuter = SklearnPipelinePermuter()
        pipeline_permuter.results = score_summary
        return pipeline_permuter

    def fit(
        self,
        X: np.ndarray,  # noqa: N803
        y: np.ndarray,
        *,
        outer_cv: BaseCrossValidator,
        inner_cv: BaseCrossValidator,
        scoring: Optional[str_t] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ):
        """Run fit for all pipeline combinations and sets of parameters.

        This function calls :func:`~biopsykit.classification.model_selection.nested_cv_param_search` for all
        Pipeline combinations and stores the results in the ``param_searches`` attribute.

        Parameters
        ----------
        X : array-like of shape (`n_samples`, `n_features`)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.
        y : array-like of shape (`n_samples`, `n_output`) or (`n_samples`,)
            Target (i.e., class labels) relative to X for classification or regression.
        outer_cv :
            Cross-validation object determining the cross-validation splitting strategy of the outer cross-validation.
        inner_cv :
            Cross-validation object determining the cross-validation splitting strategy of the hyperparameter search.
        scoring : str, optional
            A str specifying the scoring metric to use for evaluation.
        use_cache : bool, optional
            ``True`` to cache fitted transformer instances of the pipeline in a caching directory
            (can be provided by the additional parameter ``cachedir_name``), ``False`` otherwise. Default: ``True``
        **kwargs :
            Additional arguments that are passed to
            :func:`~biopsykit.classification.model_selection.nested_cv_parameter_search` and the hyperparameter search
            class instance (e.g., :class:`~sklearn.model_selection.GridSearchCV` or
            :class:`~sklearn.model_selection.RandomizedSearchCV`).

        """
        return self._fit(X=X, y=y, outer_cv=outer_cv, inner_cv=inner_cv, scoring=scoring, use_cache=use_cache, **kwargs)

    def _fit(  # noqa: PLR0912, C901
        self,
        *,
        X: np.ndarray,  # noqa: N803
        y: np.ndarray,
        outer_cv: BaseCrossValidator,
        inner_cv: BaseCrossValidator,
        save_intermediate: Optional[bool] = False,
        file_path: Optional[path_t] = None,
        scoring: Optional[str_t] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ):
        self.results = None

        if len(self.model_combinations) == 0:
            raise ValueError("No model combinations specified. Please specify at least one model combination.")

        kwargs.setdefault("n_jobs", -1)
        kwargs.setdefault("verbose", 1)
        kwargs.setdefault("error_score", "raise")

        # Create a temporary folder to store the transformers of the pipeline
        location = kwargs.pop("cachedir_name", "cachedir")
        memory = None
        if use_cache:
            memory = Memory(location=location, verbose=0)

        if scoring is None:
            scoring = "accuracy"
        self.scoring = scoring

        refit = kwargs.get("refit")
        if refit is None:
            refit = scoring
        self.refit = refit

        for model_combination in tqdm(self.model_combinations, desc="Pipeline Combinations"):
            if model_combination in self.param_searches:
                print(f"Skipping {model_combination} since this combination was already fitted!")
                # continue if we already tried this combination
                continue

            pipeline_params = [(m, self.params[k[1]]) for m, k in zip(self.models.keys(), model_combination)]
            pipeline_params = list(filter(lambda p: p[1] is not None, pipeline_params))
            pipeline_params = [(m, k_new) for m, k in pipeline_params for k_new in k if k is not None]

            cats = {p[0] for p in pipeline_params}
            pipeline_params = [list(filter(lambda p, c=cat: p[0] == c, pipeline_params)) for cat in cats]
            pipeline_params = list(product(*pipeline_params))

            pipeline_params = [
                tuple({f"{step[0]}__{k}": v for k, v in step[1].items()} for step in combi) for combi in pipeline_params
            ]
            pipeline_params = [{k: v for x in param for k, v in x.items()} for param in pipeline_params]

            if kwargs["verbose"] >= 1:
                print(
                    f"### Running hyperparameter search for pipeline: "
                    f"{model_combination} with {len(pipeline_params)} parameter grid(s):"
                )

            for j, param_dict in enumerate(pipeline_params):
                hyper_search_params = self.hyper_search_dict[model_combination[-1][1]]
                model_cls = [(step, self.models[step][m]) for step, m in model_combination]
                for i in range(len(model_cls)):
                    if isinstance(model_cls[i][1], BaseEstimator):
                        model_cls[i] = (model_cls[i][0], clone(model_cls[i][1]))

                pipeline = Pipeline(model_cls, memory=memory)
                if kwargs["verbose"] >= 1:
                    print(f"Parameter grid #{j} ({hyper_search_params}): {param_dict}")

                result_dict = nested_cv_param_search(
                    X,
                    y,
                    param_dict=param_dict,
                    pipeline=pipeline,
                    outer_cv=outer_cv,
                    inner_cv=inner_cv,
                    scoring=scoring,
                    hyper_search_params=hyper_search_params,
                    random_state=self.random_state,
                    **kwargs,
                )

                self.param_searches[model_combination] = result_dict
                if kwargs["verbose"] >= 1:
                    print("")

            if save_intermediate:
                # Save intermediate results to file
                self.to_pickle(file_path)
            if kwargs["verbose"] >= 1:
                print("")

        if use_cache:
            # Delete the temporary cache before exiting
            memory.clear(warn=False)
            rmtree(location)

    def fit_and_save_intermediate(
        self,
        X: np.ndarray,  # noqa: N803
        y: np.ndarray,
        *,
        outer_cv: BaseCrossValidator,
        inner_cv: BaseCrossValidator,
        file_path: path_t,
        scoring: Optional[str_t] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ):
        """Run fit for all pipeline combinations and sets of parameters and save intermediate results to file.

        This function calls :func:`~biopsykit.classification.model_selection.nested_cv_param_search` for all
        Pipeline combinations and stores the results in the ``param_searches`` attribute. After each model combination,
        the results are saved to a pickle file.

        Parameters
        ----------
        X : array-like of shape (`n_samples`, `n_features`)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.
        y : array-like of shape (`n_samples`, `n_output`) or (`n_samples`,)
            Target (i.e., class labels) relative to X for classification or regression.
        outer_cv :
            Cross-validation object determining the cross-validation splitting strategy of the outer cross-validation.
        inner_cv :
            Cross-validation object determining the cross-validation splitting strategy of the hyperparameter search.
        file_path : :class:`pathlib.Path` or str
            path to pickle file
        scoring : str, optional
            A str specifying the scoring metric to use for evaluation.
        use_cache : bool, optional
            ``True`` to cache fitted transformer instances of the pipeline in a caching directory
            (can be provided by the additional parameter ``cachedir_name``), ``False`` otherwise. Default: ``True``
        **kwargs :
            Additional arguments that are passed to
            :func:`~biopsykit.classification.model_selection.nested_cv_parameter_search` and the hyperparameter search
            class instance (e.g., :class:`~sklearn.model_selection.GridSearchCV` or
            :class:`~sklearn.model_selection.RandomizedSearchCV`).

        """
        return self._fit(
            X=X,
            y=y,
            outer_cv=outer_cv,
            inner_cv=inner_cv,
            save_intermediate=True,
            file_path=file_path,
            scoring=scoring,
            use_cache=use_cache,
            **kwargs,
        )

    @functools.lru_cache(maxsize=5)  # noqa: B019
    def pipeline_score_results(self) -> pd.DataFrame:
        """Return parameter search results for each pipeline combination.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with parameter search results for each pipeline combination

        """
        if self._results_set:
            return self.results
        if len(self.param_searches) == 0:
            raise AttributeError(
                "No results available because pipelines were not fitted! Call `SklearnPipelinePermuter.fit()` first."
            )

        gs_param_list = []
        for param, gs in self.param_searches.items():
            dict_folds = {}
            for i, res in enumerate(gs["cv_results"]):
                df_res = pd.DataFrame(res)
                df_res = df_res.drop(columns=df_res.filter(like="time").columns)
                df_res.index.name = "parameter_combination_id"
                # note: the best_estimator from the fold does not necessarily correspond to the pipeline returned
                # from best_pipeline() because the best_pipeline is determined by averaging over all folds and
                # can hence have a different set of hyperparameters
                # best_estimator = gs["best_estimator"][i]
                # df_res["best_estimator"] = _PipelineWrapper(best_estimator)
                dict_folds[i] = df_res
            param_dict = {f"pipeline_{key}": val for key, val in param}
            df_gs = pd.concat(dict_folds, names=["outer_fold"])
            df_gs[list(param_dict.keys())] = [list(param_dict.values())] * len(df_gs)
            df_gs = df_gs.set_index(list(df_gs.filter(like="pipeline").columns), append=True)
            # reorder levels so that pipeline steps are the first index levels, then the hyperparameter
            # combination id, then the outer_fold id
            gs_param_list.append(df_gs)

        df_summary = pd.concat(gs_param_list)
        df_summary = df_summary.infer_objects()

        df_summary = df_summary.reorder_levels(
            df_summary.index.names[2:] + [df_summary.index.names[1]] + [df_summary.index.names[0]]
        )
        self.results = df_summary.sort_index().sort_index(axis=1)
        return self.results

    def metric_summary(
        self, additional_metrics: Optional[str_t] = None, pos_label: Optional[str] = None
    ) -> pd.DataFrame:
        """Return summary with all performance metrics for the `best-performing estimator` of each pipeline combination.

        The `best-performing estimator` for each pipeline combination is the `best_estimator_` that
        :class:`~sklearn.model_selection.GridSearchCV` returns for each outer fold, i.e. the pipeline which yielded
        the highest average test score (over all inner folds).

        Parameters
        ----------
        additional_metrics : str or list of str, optional
            additional metrics to compute. Default: ``None``. Available metrics can be found in scikit-learn's
            `metrics and scoring <https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_
            module.
        pos_label : str, optional
            positive label for binary classification, must be specified if `additional_metrics` is specified.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with performance metric summary the `best estimator` of each pipeline combination.

        """
        if len(self.param_searches) == 0:
            raise AttributeError(
                "No results available because pipelines were not fitted! Call `SklearnPipelinePermuter.fit()` first."
            )
        list_metric_summary = []
        for param_key, param_value in self.param_searches.items():
            param_dict = {f"pipeline_{key}": val for key, val in param_key}
            conf_matrix = np.sum(param_value["conf_matrix"], axis=0)
            true_labels = np.array(param_value["true_labels"], dtype="object")
            predicted_labels = np.array(param_value["predicted_labels"], dtype="object")
            train_indices = np.array(param_value["train_indices"], dtype="object")
            test_indices = np.array(param_value["test_indices"], dtype="object")
            df_metric = pd.DataFrame(param_dict, index=[0])

            df_metric["conf_matrix"] = [list(conf_matrix.flatten())]
            df_metric["conf_matrix_folds"] = [[cm.flatten() for cm in param_value["conf_matrix"]]]
            df_metric["true_labels"] = [np.concatenate(true_labels)]
            df_metric["true_labels_folds"] = [true_labels]
            df_metric["predicted_labels"] = [np.concatenate(predicted_labels)]
            df_metric["predicted_labels_folds"] = [predicted_labels]
            df_metric["train_indices"] = [np.concatenate(train_indices)]
            df_metric["train_indices_folds"] = [train_indices]
            df_metric["test_indices"] = [np.concatenate(test_indices)]
            df_metric["test_indices_folds"] = [test_indices]

            scoring = self.scoring
            if isinstance(scoring, str):
                scoring = [scoring]
            for score_key in scoring:
                key = f"test_{score_key}"
                test_scores = self.param_searches[param_key][key]
                df_metric[f"mean_{key}"] = np.mean(test_scores)
                df_metric[f"std_{key}"] = np.std(test_scores)
                df_metric[[f"{key}_fold_{i}" for i in range(len(test_scores))]] = list(test_scores)

            df_metric = df_metric.set_index(list(df_metric.columns)[: len(param_dict)])
            list_metric_summary.append(df_metric)

        metric_summary = pd.concat(list_metric_summary)

        if additional_metrics is not None:
            metric_summary = self.compute_additional_metrics(
                metric_summary, metrics=additional_metrics, pos_label=pos_label
            )

        return metric_summary

    def export_pipeline_score_results(self, file_path: path_t) -> None:
        """Export pipeline score results as csv file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to export

        """
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".csv")
        self.results.to_csv(file_path)

    def export_metric_summary(self, file_path: path_t) -> None:
        """Export performance metric summary as csv file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to export

        """
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".csv")
        self.metric_summary().to_csv(file_path, sep=";")

    def best_estimator_summary(self) -> pd.DataFrame:
        """Return a dataframe with the `best estimator` instances of all pipeline combinations for each fold.

        Each entry of the dataframe is a list of :class:`~sklearn.pipeline.Pipeline` objects whe returned the .

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with `best estimator` instances

        """
        best_estimator_list = []
        for param_key, param_value in self.param_searches.items():
            param_dict = {f"pipeline_{key}": val for key, val in param_key}
            df_be = pd.DataFrame(param_dict, index=[0])
            df_be["best_estimator"] = _PipelineWrapper(param_value["best_estimator"])
            df_be = df_be.set_index(list(df_be.columns)[:-1])
            best_estimator_list.append(df_be)

        return pd.concat(best_estimator_list)

    @functools.lru_cache(maxsize=5)  # noqa: B019
    def mean_pipeline_score_results(self) -> pd.DataFrame:
        """Compute mean score results for each pipeline combination and hyperparameter combination.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with mean score results for each pipeline combination and each parameter combination,
            sorted by the highest mean score.

        Notes
        -----
        The pipeline with the highest "mean over the mean test scores" does not necessarily correspond to the
        best-performing pipeline as returned by
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.metric_summary` or
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.best_estimator_summary` because the
        best-performing pipelines are determined by averaging the `best_estimator` instances, as determined by
        `scikit-learn` over all folds. Hence, all `best_estimator` instances can have a **different** set of
        hyperparameters.

        This function should only be used if you want to gain a deeper understanding of the different hyperparameter
        combinations and their performance. If you want to get the best-performing pipeline(s) to report in a paper,
        use :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.metric_summary` or
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.best_estimator_summary` instead.

        """
        score_results = self.pipeline_score_results()
        cols_scoring = [col for col in score_results.columns if any(c for c in ["test", "train"] if c in col)]
        score_summary_mean = (
            score_results[cols_scoring]
            .groupby(score_results.index.names[:-1])
            .agg(["mean", "std"])
            .sort_values(by=(f"mean_test_{self.refit}", "mean"), ascending=False)
        )
        return score_summary_mean

    def best_hyperparameter_pipeline(self) -> pd.DataFrame:
        """Return the evaluation results for the pipeline with the best-performing hyperparameter set.

        This returns the pipeline with the **unique** hyperparameter combination that achieved
        the highest mean score over all outer folds.

        Notes
        -----
        This `best pipeline` does not necessarily correspond to the overall best-performing pipeline as returned by
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.metric_summary` or
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.best_estimator_summary` because the
        best-performing pipelines are determined by averaging the `best_estimator` instances, as determined by
        `scikit-learn` over all folds. Hence, all `best_estimator` instances can have a **different** set of
        hyperparameters. This function returns the pipeline with the **unique** hyperparameter combination that
        achieved the highest mean score over all outer folds.

        This function should only be used if you want to gain a deeper understanding of the different hyperparameter
        combinations and their performance. If you want to get the best-performing pipeline(s) to report in a paper,
        use :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.metric_summary` or
        :func:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.best_estimator_summary` instead.


        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with the evaluation results of the best pipeline over all outer folds

        """
        score_summary = self.pipeline_score_results()
        score_summary_mean = self.mean_pipeline_score_results()
        return score_summary.loc[score_summary_mean.index[0]].dropna(how="all", axis=1)

    @staticmethod
    def _check_missing_params(
        model_dict: dict[str, dict[str, BaseEstimator]],
        param_dict: dict[str, Optional[Union[Sequence[dict[str, Any]], dict[str, Any]]]],
    ):
        for _category, estimator_dict in model_dict.items():
            if not set(estimator_dict.keys()).issubset(set(param_dict.keys())):
                missing_params = list(set(estimator_dict.keys()) - set(param_dict.keys()))
                raise ValueError(f"Some estimators are missing parameters: {missing_params}")

    def metric_summary_to_latex(
        self,
        data: Optional[pd.DataFrame] = None,
        metrics: Optional[Sequence[str]] = None,
        pipeline_steps: Optional[Sequence[str]] = None,
        si_table_format: Optional[str] = None,
        highlight_best: Optional[Union[str, bool]] = None,
        **kwargs,
    ) -> str:
        """Return a latex table with the performance metrics of the pipeline combinations.

        By default, this function uses the attribute of the ``SklearnPipelinePermuter`` instance.
        If the ``data`` parameter is set, the function uses the dataframe passed as argument.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`, optional
            dataframe with performance metrics if custom data should be used or ``None`` to use the attribute of the
            ``SklearnPipelinePermuter`` instance. Default: ``None``
        metrics : list of str, optional
            list of metrics to include in the table or ``None`` to use all available metrics in the dataframe.
            Default: ``None``
        pipeline_steps : list of str, optional
            list of pipeline steps to include in the table index or ``None`` to show all available pipeline steps
            as table index. Default: ``None``
        si_table_format : str, optional
            table format for the ``siunitx`` package or ``None`` to use the default format. Default: ``None``
        highlight_best : bool or str, optional
            Whether to highlight the pipeline with the best value in each column or not.
            *  If ``highlight_best`` is a boolean, the best pipeline is highlighted in each column.
            *  If ``highlight_best`` is a string, the best pipeline is highlighted in the column with the name
        **kwargs
            additional keyword arguments passed to :func:`~pandas.DataFrame.to_latex`

        """
        kwargs.setdefault("clines", "skip-last;data")
        kwargs.setdefault("hrules", True)
        kwargs.setdefault("position", "ht!")
        kwargs.setdefault("position_float", "centering")
        kwargs.setdefault("siunitx", True)
        if si_table_format is None:
            si_table_format = "table-format = 2.1(2)"

        if data is None:
            data = self.metric_summary()
        metric_summary = data.copy()

        if pipeline_steps is None:
            if isinstance(metric_summary.index, pd.MultiIndex):
                pipeline_steps = list(metric_summary.index.names)
            else:
                pipeline_steps = [metric_summary.index.name]

        if metrics is None:
            metrics = metric_summary.filter(like="mean_test").columns
            # extract metric names
            metrics = [m.split("_")[-1] for m in metrics]

        levels_to_drop = [step for step in metric_summary.index.names if step not in pipeline_steps]
        metric_summary = metric_summary.droplevel(levels_to_drop)
        metric_summary = metric_summary.rename(index=clf_map)

        list_metric_summary = []
        for metric in metrics:
            list_metric_summary.append(metric_summary.filter(regex=f"(mean|std)_test_{metric}"))

        metric_summary = pd.concat(list_metric_summary, axis=1)

        # convert to percent
        metric_summary = metric_summary * 100
        metric_summary_export = metric_summary.copy()

        for metric in metrics:
            mean_test = f"mean_test_{metric}"
            std_test = f"std_test_{metric}"
            m_sd = metric_summary_export.apply(
                lambda x, m_t=mean_test, std_t=std_test: rf"{x[m_t]:.1f}({x[std_t]:.1f})", axis=1
            )
            metric_summary_export = metric_summary_export.assign(**{metric: m_sd})
        metric_summary_export = metric_summary_export[metrics].copy()

        if isinstance(metric_summary_export.index, pd.MultiIndex):
            metric_summary_export.index = metric_summary_export.index.rename(pipeline_step_map)
        metric_summary_export = metric_summary_export.rename(columns=metric_map)

        kwargs.setdefault("column_format", self._format_latex_column_format(metric_summary_export))

        styler = metric_summary_export.style
        styler = self._highlight_best(metric_summary, styler, highlight_best, metric_summary_export)

        metric_summary_tex = styler.to_latex(**kwargs)
        metric_summary_tex = self._apply_latex_code_correction(metric_summary_tex, si_table_format)
        return metric_summary_tex

    @staticmethod
    def _format_latex_column_format(data: pd.DataFrame):
        column_format = "l" * data.index.nlevels
        if isinstance(data.columns, pd.MultiIndex):
            ncols = len(data.columns)
            ncols_last_level = len(data.columns.get_level_values(-1).unique())
            column_format += ("S" * ncols_last_level + "|") * (ncols // ncols_last_level)
            # remove the last "|"
            column_format = column_format[:-1]
        else:
            column_format += "S" * len(data.columns)
        return column_format

    @staticmethod
    def _apply_latex_code_correction(table: str, si_table_format: str) -> str:
        if si_table_format is not None:
            table = re.sub(r"(\\begin\{tabular\})", r"\\sisetup{" + si_table_format + r"}\n\n\1", table)
        return table

    def to_pickle(self, file_path: path_t) -> None:
        """Export the current instance as a pickle file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to export

        """
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".pkl")
        with file_path.open(mode="wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def from_pickle(file_path: path_t) -> "SklearnPipelinePermuter":
        """Import a ``SklearnPipelinePermuter`` instance from a pickle file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to import

        Returns
        -------
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
            ``SklearnPipelinePermuter`` instance

        """
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".pkl")
        with file_path.open(mode="rb") as f:
            return pickle.load(f)

    def update_permuter(
        self,
        model_dict: Optional[dict[str, dict[str, BaseEstimator]]] = None,
        param_dict: Optional[dict[str, Any]] = None,
        hyper_search_dict: Optional[dict[str, dict[str, Any]]] = None,
    ) -> Self:
        """Update the ``SklearnPipelinePermuter`` instance with new model and parameter dictionaries.

        Parameters
        ----------
        model_dict : dict, optional
            dictionary with model classes for each pipeline step
        param_dict : dict, optional
            dictionary with parameter grids for each pipeline step
        hyper_search_dict : dict, optional
            dictionary with hyperparameter search settings for each estimator

        Returns
        -------
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
            updated ``SklearnPipelinePermuter`` instance

        """
        permuter = SklearnPipelinePermuter(model_dict, param_dict, hyper_search_dict)
        return SklearnPipelinePermuter._merge_permuter_params(self, permuter)

    @classmethod
    def _merge_permuter_params(cls, permuter_01: Self, permuter_02: Self):
        # merge model dicts
        permuter_01.models = merge_nested_dicts(permuter_01.models, permuter_02.models)
        # merge hyperparameter search dicts
        permuter_01.hyper_search_dict = merge_nested_dicts(permuter_01.hyper_search_dict, permuter_02.hyper_search_dict)

        # merge hyperparameter dicts
        permuter_01.param_searches = merge_nested_dicts(permuter_01.param_searches, permuter_02.param_searches)
        permuter_01.params = merge_nested_dicts(permuter_01.params, permuter_02.params)

        # merge model combinations
        permuter_01.model_combinations += permuter_02.model_combinations
        permuter_01.model_combinations = list(set(permuter_01.model_combinations))

        return permuter_01

    @classmethod
    def merge_permuter_instances(cls, permuter: Union[Sequence[Self], Sequence[path_t]]) -> Self:
        """Merge two (or more) :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter` instances.

        This function expects at least two ``SklearnPipelinePermuter`` instances to merge. The function first performs
        a deep copy of the first instance and then merges all attributes of the remaining ``permuter`` instance with
        the copy. The ``permuter`` instances passed to this function are not modified.

        Parameters
        ----------
        permuter : list of :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter` instances or
                    list of file paths to pickled `SklearnPipelinePermuter` instances

        Returns
        -------
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
            merged ``SklearnPipelinePermuter`` instance

        """
        # ensure that permuter contains at least two instances
        if len(permuter) < 2:
            raise ValueError("At least two SklearnPipelinePermuter instances must be passed to this function.")

        if all(isinstance(p, (str, Path)) for p in permuter):
            permuter = [cls.from_pickle(p) for p in permuter]

        # make deep copy of first instance
        base_permuter = deepcopy(permuter[0])

        for p in permuter[1:]:
            if base_permuter.scoring != p.scoring:
                raise ValueError(
                    f"Cannot merge permuter instances with different scoring functions: "
                    f"{base_permuter.scoring} vs. {p.scoring}"
                )
            if base_permuter.refit != p.refit:
                raise ValueError(
                    f"Cannot merge permuter instances with different refit options: {base_permuter.refit} vs. {p.refit}"
                )

            SklearnPipelinePermuter._merge_permuter_params(base_permuter, p)

            # merge results dataframes
            results_concat = pd.concat([base_permuter.results, p.results], axis=0)
            param_cols = list(results_concat.filter(like="param_").columns)
            # drop duplicate parameter combinations in results
            results_concat = results_concat.reset_index("outer_fold").drop_duplicates(
                subset=["outer_fold", *param_cols]
            )
            results_concat = results_concat.set_index("outer_fold", append=True)
            base_permuter.results = results_concat

        return base_permuter

    @staticmethod
    def _apply_score(row: pd.Series, score_func, pos_label: str):
        true_labels_folds = row[0]
        predicted_labels_folds = row[1]

        kwargs = {}
        params = signature(score_func).parameters
        if "pos_label" in params:
            kwargs["pos_label"] = pos_label
        if "zero_division" in params:
            kwargs["zero_division"] = 0

        scores = [
            score_func(true_labels, predicted_labels, **kwargs)
            for true_labels, predicted_labels in zip(true_labels_folds, predicted_labels_folds)
        ]
        return pd.Series(scores)

    def compute_additional_metrics(self, metric_summary: pd.DataFrame, metrics: str_t, pos_label: str) -> pd.DataFrame:
        """Compute additional classification metrics.

        Parameters
        ----------
        metric_summary : :class:`~pandas.DataFrame`
            metric summary from :meth:`~biopsykit.classification.model_selection.SklearnPipelinePermuter.metric_summary`
        metrics : str or list of str
            metric(s) to compute
        pos_label : str
            positive label for binary classification

        Returns
        -------
        :class:`~pandas.DataFrame`
            metric summary with additional metrics computed

        """
        if isinstance(metrics, str):
            metrics = [metrics]
        metric_slice = metric_summary[["true_labels_folds", "predicted_labels_folds"]].copy()
        metric_out = {}
        score_funcs = dict(getmembers(sklearn.metrics))
        for metric in metrics:
            if metric.endswith("_score"):
                score_name = metric
                # strip '_score' suffix from metric name for column name
                metric = metric.replace("_score", "")  # noqa: PLW2901
            else:
                # name for calling sklearn metric function
                score_name = metric + "_score"

            if score_name in score_funcs:
                score_func = score_funcs[score_name]
            else:
                raise ValueError(f"Metric '{metric}' not found.")

            metric_out[metric] = metric_slice.apply(self._apply_score, args=(score_func, pos_label), axis=1)
        metric_out = pd.concat(metric_out, names=["score", "folds"], axis=1)

        metric_out = metric_out.stack(["score", "folds"])
        metric_out = metric_out.groupby(metric_out.index.names[:-1]).agg(
            [("mean", lambda x: np.mean(x)), ("std", lambda x: np.std(x))]
        )

        metric_out = metric_out.unstack("score").sort_index(axis=1, level="score")
        metric_out.columns = metric_out.columns.map("_test_".join)
        metric_summary = metric_summary.join(metric_out)

        # resort columns so that all "mean_test_*" and "std_test_*" columns are at the end
        cols = list(metric_summary.filter(regex="^(?!mean_test_|std_test_).*$").columns)
        cols += list(metric_summary.filter(regex="^(mean_test_|std_test_).*$").columns)
        metric_summary = metric_summary[cols]

        return metric_summary

    @staticmethod
    def _highlight_best(metric_summary, styler, highlight_best, metric_summary_export):
        if isinstance(highlight_best, str):
            max_metric = metric_summary[f"mean_test_{highlight_best}"].idxmax()
            # get index of max metric
            max_metric = metric_summary_export.index.get_loc(max_metric)
            styler = styler.highlight_max(subset=metric_map[highlight_best], props="bfseries: ;")
            # get maximum of metric_summary
            # make index bold
            styler = styler.apply_index(lambda x: np.where(x.index == max_metric, "bfseries: ;", ""))
        elif isinstance(highlight_best, bool) and highlight_best:
            styler = styler.highlight_max(props="bfseries: ;")
        return styler



"""Functions to analyze classification results."""

from biopsykit.classification.analysis._analysis import (
    metric_summary_to_latex,
    plot_conf_matrix,
    plot_conf_matrix_proba,
    predict_proba_from_estimator,
    predictions_as_df,
)

__all__ = [
    "metric_summary_to_latex",
    "plot_conf_matrix",
    "plot_conf_matrix_proba",
    "predict_proba_from_estimator",
    "predictions_as_df",
]



"""I/O functions for files related to EEG processing."""

import warnings
from typing import Optional

import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_has_columns
from biopsykit.utils._types_internal import path_t

__all__ = ["MuseDataset", "load_eeg_raw_muse"]


class MuseDataset:
    """Class for loading and processing EEG data from the Muse EEG wearable headband."""

    SAMPLING_RATE_HZ = 250.0
    _start_time_unix: pd.Timestamp
    _tz: str
    _data: pd.DataFrame
    _sampling_rate: int

    def __init__(self, data: pd.DataFrame, tz: Optional[str] = None):
        """Get new Dataset instance.

        .. note::
            Usually you shouldn't use this init directly.
            Use the provided `from_csv_file` constructor to handle loading recorded Muse data from a CSV file.

        Parameters
        ----------
        data : :class:`pandas.DataFrame`
            DataFrame containing the raw EEG data.
        tz : str, optional
            Timezone of the recording, if present or ``None`` if no timezone is available.

        """
        self._data = data
        self._tz = tz

    @classmethod
    def from_csv_file(cls, file_path: path_t, tz: Optional[str] = None):
        """Load Muse data from a CSV file.

        Parameters
        ----------
        file_path : str or :class:`pathlib.Path`
            Path to the CSV file.
        tz : str, optional
            Timezone of the recording, if present or ``None`` if no timezone is available.

        Returns
        -------
        :class:`~biopsykit.io.eeg.MuseDataset`
            Dataset instance containing the loaded data.

        """
        data = pd.read_csv(file_path)
        _assert_has_columns(data, [["timestamps", "TP9", "AF7", "AF8", "TP10"]])
        # drop the AUX column, if present
        data = data.drop(columns="Right AUX", errors="ignore")
        return cls(data, tz=tz)

    @property
    def sampling_rate_hz(self):
        """Return the sampling rate of the EEG data in Hz."""
        return self.SAMPLING_RATE_HZ

    @property
    def timezone(self) -> str:
        """Return the timezone of the recording."""
        return self._tz

    def data_as_df(self, index: Optional[str] = None) -> pd.DataFrame:
        """Return all data as one combined :class:`pandas.DataFrame`.

        Parameters
        ----------
        index : str, optional
            Specify which index should be used for the dataset. The options are:
            * "time": For the time in seconds since the first sample
            * "utc": For the utc time stamp of each sample
            * "utc_datetime": for a pandas DateTime index in UTC time
            * "local_datetime": for a pandas DateTime index in the timezone set for the session
            * None: For a simple index (0...N)

        Returns
        -------
        :class:`pandas.DataFrame`
            DataFrame containing the raw EEG data.

        """
        data = self._add_index(self._data, index)
        return data

    def _add_index(self, data: pd.DataFrame, index: str) -> pd.DataFrame:
        index_names = {
            None: "n_samples",
            "time": "t",
            "utc": "utc",
            "utc_datetime": "date",
            "local_datetime": f"date ({self.timezone})",
        }
        if index and index not in index_names:
            raise ValueError(f"Supplied value for index ({index}) is not allowed. Allowed values: {index_names.keys()}")
        data = data.set_index("timestamps")

        index_name = index_names[index]
        data.index.name = index_name

        if index == "utc":
            return data
        if index == "time":
            data.index -= data.index[0]
            return data
        if index is None:
            data = data.reset_index(drop=True)
            data.index.name = index_name
            return data

        # convert counter to pandas datetime index
        data.index = pd.to_datetime(data.index, unit="s").tz_localize("UTC")

        if index == "local_datetime":
            data.index = data.index.tz_convert(self.timezone)

        return data


def load_eeg_raw_muse(file_path: path_t, tz: Optional[str] = "Europe/Berlin") -> tuple[pd.DataFrame, float]:
    """Load a csv file with raw EEG data recorded by the Muse headband.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to file
    tz : str, optional
        timezone of the recording, if present or ``None`` if no timezone is available. Default: "Europe/Berlin"

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with raw EEG data
    fs : float
        sampling rate

    Raises
    ------
    ValidationError
        if file specified by ``file_path`` does not contain the required `timestamp` column as well as the
        EEG channel columns

    """
    # raise deprecation warning
    warnings.warn(
        "The function `load_eeg_raw_muse` is deprecated and will be removed in a future version. "
        "Use the `MuseDataset` class instead.",
        category=DeprecationWarning,
    )
    fs = 250.0
    data = pd.read_csv(file_path)

    _assert_has_columns(data, [["timestamps", "TP9", "AF7", "AF8", "TP10"]])
    # convert timestamps to datetime object, set as dataframe index
    data["timestamps"] = pd.to_datetime(data["timestamps"], unit="s")
    data = data.set_index("timestamps")
    # convert timestamp from UTC into the correct time zone
    data = data.tz_localize("UTC").tz_convert(tz)
    # drop the AUX column, if present
    data = data.drop(columns="Right AUX", errors="ignore")
    return data, fs



"""Functions for processing saliva data and computing established features (AUC, slope, maximum increase, ...)."""

import warnings
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import pandas as pd
import scipy.stats as ss

from biopsykit.saliva.utils import (
    _check_sample_times,
    _get_group_cols,
    _get_saliva_idx_labels,
    _get_sample_times,
    _remove_s0,
)
from biopsykit.utils.dtypes import (
    SalivaFeatureDataFrame,
    SalivaMeanSeDataFrame,
    SalivaRawDataFrame,
    _SalivaFeatureDataFrame,
    _SalivaMeanSeDataFrame,
    is_saliva_feature_dataframe,
    is_saliva_mean_se_dataframe,
    is_saliva_raw_dataframe,
)
from biopsykit.utils.exceptions import DataFrameTransformationError
from biopsykit.utils.functions import se


def max_value(
    data: SalivaRawDataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    remove_s0: Optional[bool] = False,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    """Compute maximum value.

    The output feature name will be ``max_val``, preceded by the name of the saliva type to allow better
    conversion into long-format later on (if desired). So e.g., for cortisol, it will be: ``cortisol_max_val``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    remove_s0 : bool, optional
        whether to remove the first saliva sample for computing maximum or not. Default: ``False``

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = max_value(
                data[saliva_col],
                saliva_type=saliva,
                remove_s0=remove_s0,
            )
        return dict_result

    if remove_s0:
        # We have a S0 sample => drop it
        data = data.drop(0, level="sample", errors="ignore")
        data = data.drop("0", level="sample", errors="ignore")
        data = data.drop("S0", level="sample", errors="ignore")

    data = data[[saliva_type]].unstack(level="sample")

    max_val = data.max(axis=1)

    out = pd.DataFrame(
        max_val,
        columns=[f"{saliva_type}_max_val"],
        index=max_val.index,
    )
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)
    return out


def initial_value(
    data: SalivaRawDataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    remove_s0: Optional[bool] = False,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    """Compute initial saliva sample.

    The output feature name will be ``ini_val``, preceded by the name of the saliva type to allow better
    conversion into long-format later on (if desired). So e.g., for cortisol, it will be: ``cortisol_ini_val``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    remove_s0 : bool, optional
        whether to remove the first saliva sample for computing initial value or not. Default: ``False``

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = initial_value(
                data[saliva_col],
                saliva_type=saliva,
                remove_s0=remove_s0,
            )
        return dict_result

    if remove_s0:
        # We have a S0 sample => drop it
        data = _remove_s0(data)

    data = data[[saliva_type]].unstack(level="sample")

    ini_val = data.iloc[:, 0]

    out = pd.DataFrame(
        ini_val.values,
        columns=[f"{saliva_type}_ini_val"],
        index=ini_val.index,
    )
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)
    return out


def max_increase(
    data: SalivaRawDataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    remove_s0: Optional[bool] = False,
    percent: Optional[bool] = False,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    """Compute maximum increase between first saliva sample and all others.

    The maximum increase (`max_inc`) is defined as the difference between the `first` sample and the maximum of
    all `subsequent` samples.

    If the first sample should be excluded from computation, e.g., because the first sample was just collected for
    controlling against high initial saliva levels, ``remove_s0`` needs to set to ``True``.

    The output is either the absolute increase or the relative increase to the first sample in percent
    (if ``percent`` is ``True``).

    The output feature name will be ``max_inc``, preceded by the name of the saliva type to allow better
    conversion into long-format later on (if desired). So e.g., for cortisol, it will be: ``cortisol_max_inc``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    remove_s0 : bool, optional
        whether to exclude the first saliva sample from computing `max_inc` or not. Default: ``False``
    percent : bool, optional
        whether to compute ``max_inc`` in percent (i.e., relative increase) or not. Default: ``False``

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = max_increase(
                data[saliva_col],
                saliva_type=saliva,
                remove_s0=remove_s0,
                percent=percent,
            )
        return dict_result

    if remove_s0:
        # We have a S0 sample => drop it
        data = _remove_s0(data)

    data = data[[saliva_type]].unstack(level="sample")

    max_inc = data.iloc[:, 1:].max(axis=1) - data.iloc[:, 0]
    if percent:
        max_inc = 100.0 * max_inc / np.abs(data.iloc[:, 0])

    out = pd.DataFrame(
        max_inc,
        columns=[f"{saliva_type}_max_inc_percent" if percent else f"{saliva_type}_max_inc"],
        index=max_inc.index,
    )
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)
    return out


def auc(
    data: SalivaRawDataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    remove_s0: Optional[bool] = False,
    compute_auc_post: Optional[bool] = False,
    sample_times: Optional[Union[np.ndarray, Sequence[int], str]] = None,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    r"""Compute area-under-the-curve (AUC) for saliva samples.

    The area-under-the-curve is computed according to Pruessner et al. (2003) using the trapezoidal rule
    (:func:`numpy.trapz`). To compute an AUC the saliva time points are required in minutes. They can either be part of
    the :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame` (`time` column) or can be supplied as extra
    parameter (``sample_times``).

    Pruessner defined two types of AUC, which are computed by default:

    * AUC with respect to `ground` (:math:`AUC_{G}`), and
    * AUC with respect to the first sample, i.e., AUC with respect to `increase` (:math:`AUC_{I}`)

    If the first sample should be excluded from computation, e.g., because the first sample was just collected for
    controlling against high initial saliva levels, ``remove_s0`` needs to set to ``True``.

    If saliva samples were collected during an acute stress task :math:`AUC_{I}` can additionally be computed
    only for the saliva values *after* the stressor by setting ``compute_auc_post`` to ``True``.

    .. note::
        For a *pre/post* stress scenario *post*-stress saliva samples are indicated by time points :math:`t \geq 0`,
        saliva sampled collected *before* start of the stressor are indicated by time points :math:`t < 0`.
        This means that a saliva sample collected at time :math:`t = 0` is defined as *right after stressor*.

    The feature names will be ``auc_g``, ``auc_i`` (and ``auc_i_post`` if ``compute_auc_post`` is ``True``),
    preceded by the name of the saliva type to allow better conversion into long-format later on (if desired).
    So e.g., for cortisol, it will be: ``cortisol_auc_g``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    remove_s0 : bool, optional
        whether to exclude the first saliva sample from computing ``auc`` or not. Default: ``False``
    compute_auc_post : bool, optional
        whether to additionally compute :math:`AUC_I` only for saliva samples *post* stressor.
        Saliva samples *post* stressor are defined as all samples with non-negative ``sample_times``.
        Default: ``False``
    sample_times: :any:`numpy.ndarray` or list, optional
        Saliva sampling times (corresponding to x-axis values for computing AUC). By default
        (``sample_times`` is ``None``) sample times are expected to be part of the dataframe (in the `time` column).
        Alternatively, sample times can be specified by passing a list or a numpy array to this argument.
        If ``sample_times`` is a 1D array, it is assumed that saliva times are the same for all subjects.
        Then, ``sample_times`` needs to have the shape (n_samples,).
        If ``sample_times`` is a 2D array, it is assumed that saliva times are individual for all subjects.
        Then, ``saliva_times`` needs to have the shape (n_subjects, n_samples).

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`

    References
    ----------
    Pruessner, J. C., Kirschbaum, C., Meinlschmid, G., & Hellhammer, D. H. (2003).
    Two formulas for computation of the area under the curve represent measures of total hormone concentration
    versus time-dependent change. Psychoneuroendocrinology, 28(7), 916-931.
    https://doi.org/10.1016/S0306-4530(02)00108-7

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = auc(
                data[saliva_col],
                saliva_type=saliva,
                remove_s0=remove_s0,
                sample_times=sample_times,
            )
        return dict_result

    sample_times = _get_sample_times(data, saliva_type, sample_times, remove_s0)
    _check_sample_times(sample_times)

    if remove_s0:
        # We have a S0 sample => drop it
        data = _remove_s0(data)

    data = data[[saliva_type]].unstack(level="sample")

    auc_data = {
        "auc_g": np.trapz(data, sample_times),
        "auc_i": np.trapz(data.sub(data.iloc[:, 0], axis=0), sample_times),
    }

    if compute_auc_post:
        auc_data = _auc_compute_auc_post(data, auc_data, sample_times)

    out = pd.DataFrame(auc_data, index=data.index).add_prefix(f"{saliva_type}_")
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)

    return _SalivaFeatureDataFrame(out)


def _auc_compute_auc_post(
    data: SalivaRawDataFrame, auc_data: dict[str, np.ndarray], sample_times: np.ndarray
) -> dict[str, np.ndarray]:
    idxs_post = None
    if sample_times.ndim == 1:
        idxs_post = np.where(sample_times >= 0)[0]
    elif sample_times.ndim == 2:
        warnings.warn(
            "Not computing `auc_i_post` values because this is only implemented if `saliva_times` "
            "are the same for all subjects."
        )
    if idxs_post is not None:
        data_post = data.iloc[:, idxs_post]
        auc_data["auc_i_post"] = np.trapz(data_post.sub(data_post.iloc[:, 0], axis=0), sample_times[idxs_post])
    return auc_data


def slope(
    data: SalivaRawDataFrame,
    sample_labels: Optional[Union[tuple, Sequence]] = None,
    sample_idx: Optional[Union[tuple[int, int], Sequence[int]]] = None,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    sample_times: Optional[Sequence[int]] = None,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    """Compute the slope between two saliva samples.

    The samples to compute the slope can either be specified by `index` (parameter `sample_idx`) [0, num_of_samples-1]
    or by `label` (parameter `sample_idx`).

    The output feature name for the slope between saliva samples with labels `label1` and `label2` will be
    ``slope<label1><label2>``, preceded by the name of the saliva type to allow better conversion into
    long-format later on (if desired). So e.g., for cortisol, it will be: ``cortisol_slopeS1S2``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    sample_labels : list or tuple
        pair of saliva sample labels to compute slope between.
        Labels correspond to the names in the `sample` column of the dataframe.
        An error will the raised if not exactly 2 samples are specified.
    sample_idx : list or tuple
        pair of saliva sample indices to compute slope between.
        An error will the raised if not exactly 2 sample are specified
    sample_times: :any:`numpy.ndarray` or list, optional
        Saliva sampling times (corresponding to x-axis values for computing slope). By default
        (``sample_times`` is ``None``) sample times are expected to be part of the dataframe (in the `time` column).
        Alternatively, sample times can be specified by passing a list or a numpy array to this argument.
        If ``sample_times`` is a 1D array, it is assumed that saliva times are the same for all subjects.
        Then, ``sample_times`` needs to have the shape (n_samples,).
        If ``sample_times`` is a 2D array, it is assumed that saliva times are individual for all subjects.
        Then, ``saliva_times`` needs to have the shape (n_subjects, n_samples).

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    IndexError
        if invalid `sample_labels` or `sample_idx` is provided
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if sample_idx is None and sample_labels is None:
        raise IndexError("Either `sample_labels` or `sample_idx` must be supplied as parameter!")

    if sample_idx is not None and sample_labels is not None:
        raise IndexError("Either `sample_labels` or `sample_idx` must be supplied as parameter, not both!")

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = slope(
                data[saliva_col],
                sample_labels=sample_labels,
                sample_idx=sample_idx,
                saliva_type=saliva,
                sample_times=sample_times,
            )
        return dict_result

    sample_times = _get_sample_times(data, saliva_type, sample_times)
    _check_sample_times(sample_times)

    data = data[[saliva_type]].unstack()

    sample_labels, sample_idx = _get_saliva_idx_labels(
        data[saliva_type].columns, sample_labels=sample_labels, sample_idx=sample_idx
    )

    out = pd.DataFrame(
        np.diff(data.iloc[:, sample_idx]) / np.diff(sample_times[..., sample_idx]),
        index=data.index,
        columns=["{}_slope{}{}".format(saliva_type, *sample_labels)],
    )
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)

    return _SalivaFeatureDataFrame(out)


def standard_features(
    data: pd.DataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    group_cols: Optional[Union[str, Sequence[str]]] = None,
    keep_index: Optional[bool] = True,
) -> Union[SalivaFeatureDataFrame, dict[str, SalivaFeatureDataFrame]]:
    """Compute a set of `standard features` on saliva data.

    The following list of features is computed:

    * ``argmax``: Argument (=index) of the maximum value
    * ``mean``: Mean value
    * ``std``: Standard deviation
    * ``skew``: Skewness
    * ``kurt``: Kurtosis

    For all features the built-in pandas functions (e.g. :meth:`pandas.DataFrame.mean`) will be used,
    except for ``argmax``, which will use numpy's function (:func:`numpy.argmax`). The functions will be applied on the
    dataframe using the `aggregate` functions from pandas (:meth:`pandas.DataFrame.agg`).

    The output feature names will be ``argmax``, ``mean``, ``std``, ``skew``, ``kurt``, preceded by the name of the
    saliva type to allow better conversion into long-format later on (if desired).
    So e.g., for cortisol, it will be: ``cortisol_argmax``.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    group_cols: str or list of str, optional
        columns to group on before applying the aggregate function. If ``group_cols`` is ``None`` (the default),
        data will be grouped on by all columns except the `sample` column. Usually, data wants to be grouped by
        `subject` (followed by `condition`, `day`, `night`, etc., if applicable).
    keep_index : bool, optional
        whether to try keeping the old index or use the new index returned by the groupby-aggregate-function.
        Keeping the old index is e.g. useful if the dataframe has a multiindex with several levels, but grouping is
        only performed on a subset of these levels. Default: ``True``

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame` or dict of such
        dataframe containing the computed features, or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
    :exc:`~biopsykit.utils.exceptions.DataFrameTransformationError`
        if ``keep_index`` is ``True``, but applying the old index fails

    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            saliva_col = [saliva]
            if "time" in data:
                saliva_col = [*saliva_col, "time"]
            dict_result[saliva] = standard_features(data[saliva_col], saliva_type=saliva)
        return dict_result

    group_cols = _get_group_cols(data, group_cols, "sample", "standard_features")

    out = (
        data[[saliva_type]]
        .groupby(group_cols)
        .agg(
            [
                np.argmax,
                np.mean,
                np.std,
                ss.skew,
                ss.kurtosis,
            ],
        )
    )
    if keep_index:
        try:
            out.index = data.unstack(level="sample").index
        except ValueError as e:
            raise DataFrameTransformationError(
                "DataFrame transformation failed: Unable to keep old dataframe index because index does not match with "
                "output data shape, possibly because 'groupby' recuded the index. "
                "Consider setting 'keep_index' to 'False'. "
                f"The exact error was:\n\n{e!s}"
            ) from e

    # drop 'saliva_type' multiindex column and add as prefix to columns to ensure consistent naming with
    # the other saliva functions
    out.columns = out.columns.droplevel(0)
    out = out.add_prefix(f"{saliva_type}_")
    out.columns.name = "saliva_feature"

    # check output
    is_saliva_feature_dataframe(out, saliva_type)

    return _SalivaFeatureDataFrame(out)


def mean_se(
    data: SalivaRawDataFrame,
    saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
    group_cols: Optional[Union[str, list[str]]] = None,
    remove_s0: Optional[bool] = False,
) -> Union[SalivaMeanSeDataFrame, dict[str, SalivaMeanSeDataFrame]]:
    """Compute mean and standard error per saliva sample.

    Parameters
    ----------
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str or list of str
        saliva type or list of saliva types to compute features on
    group_cols: str or list of str, optional
        columns to group on before computing mean and se. If ``group_cols`` is ``None`` (the default),
        data will be grouped on by all columns except the `sample` column. Usually, data wants to be grouped by
        `subject` (followed by `condition`, `day`, `night`, etc., if applicable).
    remove_s0 : bool, optional
        whether to exclude the first saliva sample from computing mean and standard error or not. Default: ``False``

    Returns
    -------
    :class:`~biopsykit.utils.dtypes.SalivaMeanSeDataFrame`
        dataframe with mean and standard error per saliva sample or a dict of such if ``saliva_type`` is a list

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`


    """
    # check input
    is_saliva_raw_dataframe(data, saliva_type)

    if isinstance(saliva_type, list):
        dict_result = {}
        for biomarker in saliva_type:
            biomarker_cols = [biomarker]
            if "time" in data:
                biomarker_cols = ["time", *biomarker_cols]
            dict_result[biomarker] = mean_se(data[biomarker_cols], saliva_type=biomarker, remove_s0=remove_s0)
        return dict_result

    if remove_s0:
        # We have a S0 sample => drop it
        data = _remove_s0(data)

    group_cols = _get_group_cols(data, group_cols, "subject", "mean_se")
    _mean_se_assert_group_cols(data, group_cols)

    if "time" in data.columns and "time" not in group_cols:
        # add 'time' column to grouper if it's in the data and wasn't added yet because otherwise
        # we would loose this column
        group_cols = [*group_cols, "time"]

    data_mean_se = data.groupby(group_cols).agg([np.mean, se])[saliva_type]
    is_saliva_mean_se_dataframe(data_mean_se)

    return _SalivaMeanSeDataFrame(data_mean_se)


def _mean_se_assert_group_cols(data: pd.DataFrame, group_cols: Sequence[str]):
    if group_cols == list(data.index.names):
        # if data should be grouped by *all* index levels we can not compute an aggregation
        raise DataFrameTransformationError(
            "Cannot compute mean and standard error on data because *all* index "
            "columns were selected as group columns, so each sample value would be one group of its own!"
        )
    if "sample" not in group_cols:
        raise DataFrameTransformationError(
            "Error computing mean and standard error on data: 'sample' index level needs to be added as group column!"
        )



"""Module for ECG event extraction."""

from biopsykit.signals.ecg.event_extraction._base_ecg_extraction import BaseEcgExtraction
from biopsykit.signals.ecg.event_extraction._q_peak_forounzafar2018 import QPeakExtractionForouzanfar2018
from biopsykit.signals.ecg.event_extraction._q_peak_martinez2004_neurokit import QPeakExtractionMartinez2004Neurokit
from biopsykit.signals.ecg.event_extraction._q_peak_scipy_findpeaks_neurokit import (
    QPeakExtractionSciPyFindPeaksNeurokit,
)
from biopsykit.signals.ecg.event_extraction._q_peak_vanlien2013 import QPeakExtractionVanLien2013

__all__ = [
    "BaseEcgExtraction",
    "QPeakExtractionForouzanfar2018",
    "QPeakExtractionMartinez2004Neurokit",
    "QPeakExtractionSciPyFindPeaksNeurokit",
    "QPeakExtractionVanLien2013",
]



import numpy as np
import pandas as pd
from tpcp import Parameter

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.ecg.event_extraction._base_ecg_extraction import BaseEcgExtraction
from biopsykit.utils.array_handling import sanitize_input_series

__all__ = ["QPeakExtractionForouzanfar2018"]

from biopsykit.utils.dtypes import (
    EcgRawDataFrame,
    HeartbeatSegmentationDataFrame,
    is_ecg_raw_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_q_peak_dataframe,
)


class QPeakExtractionForouzanfar2018(BaseEcgExtraction, CanHandleMissingEventsMixin):
    r"""Q-peak extraction algorithm by Forouzanfar et al. (2018) [1]_.

    This algorithm detects the Q-peak of an ECG signal based on the last sample before the R-peak that is below a
    certain threshold (:math:`-1.2 \cdot R_peak/f_s`), where R_peak is the amplitude of the R-peak and f_s is the
    sampling frequency of the ECG signal.

    In this implementation, a *scaling_factor* is used instead of *f_s* since the sampling rates of the datasets
    can differ from the original publication (2000 Hz). Thus, the scaling_factor is set to the sampling rate of the
    original publication (2000 Hz) by default.


    References
    ----------
    .. [1] Forouzanfar, M., Baker, F. C., De Zambotti, M., McCall, C., Giovangrandi, L., & Kovacs, G. T. A. (2018).
        Toward a better noninvasive assessment of preejection period: A novel automatic algorithm for B-point detection
        and correction on thoracic impedance cardiogram. Psychophysiology, 55(8), e13072.
        https://doi.org/10.1111/psyp.13072


    """

    scaling_factor: Parameter[float]

    def __init__(self, scaling_factor: float = 2000, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new ``QPeakExtractionForouzanfar2018`` algorithm instance.

        Parameters
        ----------
        scaling_factor : float, optional
            Scaling factor for the threshold used to detect the Q-peak. Default: 2000
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle missing data in the input dataframes. Default: "warn"
        """
        super().__init__(handle_missing_events=handle_missing_events)
        self.scaling_factor = scaling_factor

    # @make_action_safe
    def extract(
        self,
        *,
        ecg: EcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        sampling_rate_hz: float,  # noqa: ARG002
    ):
        """Extract Q-peaks from given ECG signal.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        ecg: :class:`~pandas.DataFrame`
            ECG signal
        heartbeats: :class:`~pandas.DataFrame`
            DataFrame containing one row per segmented heartbeat, each row contains start, end, and R-peak
            location (in samples from beginning of signal) of that heartbeat, index functions as id of heartbeat
        sampling_rate_hz: float
            Sampling rate of ECG signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        self._check_valid_missing_handling()
        is_ecg_raw_dataframe(ecg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        ecg = sanitize_input_series(ecg, name="ecg")
        ecg = ecg.squeeze()

        # result df
        q_peaks = pd.DataFrame(index=heartbeats.index, columns=["q_peak_sample", "nan_reason"])

        # search Q-peak for each heartbeat of the given signal
        for idx, data in heartbeats.iterrows():
            heartbeat_start = data["start_sample"]
            r_peak_sample = data["r_peak_sample"]

            # set an individual threshold for detecting the Q-peaks based on the R-peak
            threshold = (-1.2 * ecg.iloc[r_peak_sample]) / self.scaling_factor

            # search for the Q-peak as the last sample before the R-peak that is below the threshold
            ecg_before_r_peak = ecg.iloc[heartbeat_start:r_peak_sample].reset_index(drop=True)
            ecg_below = np.where(ecg_before_r_peak < threshold)[0]

            if len(ecg_below) == 0:
                q_peaks.loc[q_peaks.index[idx], "q_peak_sample"] = np.nan
                q_peaks.loc[q_peaks.index[idx], "nan_reason"] = "no_value_below_threshold"
                continue

            q_peak_sample = heartbeat_start + ecg_below[-1]
            q_peaks.loc[q_peaks.index[idx], "q_peak_sample"] = q_peak_sample

        q_peaks = q_peaks.astype({"q_peak_sample": "Int64", "nan_reason": "object"})
        is_q_peak_dataframe(q_peaks)

        self.points_ = q_peaks
        return self
